# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from __future__ import annotations
import abc
from typing import (
    List,
    Sequence,
    Optional,
    Mapping,
    MutableMapping,
    MutableSet,
    Tuple,
    Collection,
    Union,
)

from pyomo.common.enums import ObjectiveSense
from pyomo.common.config import ConfigDict, ConfigValue, document_configdict
from pyomo.core.base.constraint import ConstraintData, Constraint
from pyomo.core.base.sos import SOSConstraintData, SOSConstraint
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData, ScalarParam
from pyomo.core.base.expression import ExpressionData
from pyomo.core.base.objective import ObjectiveData, Objective
from pyomo.core.base.block import BlockData, Block
from pyomo.core.base.suffix import Suffix
from pyomo.core.base.component import ActiveComponent
from pyomo.core.expr.numeric_expr import NumericValue
from pyomo.core.expr.relational_expr import RelationalExpression
from pyomo.common.collections import (
    ComponentMap,
    ComponentSet,
    OrderedSet,
    DefaultComponentMap,
)
from pyomo.common.gc_manager import PauseGC
from pyomo.common.timing import HierarchicalTimer
from pyomo.contrib.solver.common.util import get_objective
from pyomo.contrib.observer.component_collector import collect_components_from_expr
from pyomo.common.numeric_types import native_numeric_types
import warnings
import enum
from collections import defaultdict

# The ModelChangeDetector is meant to be used to automatically identify changes
# in a Pyomo model or block. Here is a list of changes that will be detected.
# Note that inactive components (e.g., constraints) are treated as "removed".
#   - new constraints that have been added to the model
#   - constraints that have been removed from the model
#   - new variables that have been detected in new or modified constraints/objectives
#   - old variables that are no longer used in any constraints/objectives
#   - new parameters that have been detected in new or modified constraints/objectives
#   - old parameters that are no longer used in any constraints/objectives
#   - new objectives that have been added to the model
#   - objectives that have been removed from the model
#   - modified constraint expressions (relies on expressions being immutable)
#   - modified objective expressions (relies on expressions being immutable)
#   - modified objective sense
#   - changes to variable bounds, domains, "fixed" flags, and values for fixed variables
#   - changes to named expressions (relies on expressions being immutable)
#   - changes to parameter values


_param_types = {ParamData, ScalarParam}


@document_configdict()
class AutoUpdateConfig(ConfigDict):
    """
    Control which parts of the model are automatically checked and/or updated upon re-solve
    """

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        if doc is None:
            doc = 'Configuration options to detect changes in model between solves'
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        #: automatically detect new/removed constraints on subsequent solves
        self.check_for_new_or_removed_constraints: bool = self.declare(
            'check_for_new_or_removed_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old constraints will not be automatically detected on 
                subsequent solves. Use False only when manually updating the change
                detector with cd.add_constraints() and cd.remove_constraints() or 
                when you are certain constraints are not being added to/removed from the 
                model.""",
            ),
        )
        #: automatically detect new/removed objectives on subsequent solves
        self.check_for_new_or_removed_objectives: bool = self.declare(
            'check_for_new_or_removed_objectives',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, new/old objectives will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver 
                with opt.add_objectives() and opt.remove_objectives() or when you 
                are certain objectives are not being added to/removed from the 
                model.""",
            ),
        )
        #: automatically detect changes to constraints on subsequent solves
        self.update_constraints: bool = self.declare(
            'update_constraints',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to existing constraints will not be automatically 
                detected on subsequent solves. This includes changes to the lower, 
                body, and upper attributes of constraints. Use False only when 
                manually updating the solver with opt.remove_constraints() and 
                opt.add_constraints() or when you are certain constraints are not 
                being modified.""",
            ),
        )
        #: automatically detect changes to variables on subsequent solves
        self.update_vars: bool = self.declare(
            'update_vars',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to existing variables will not be automatically 
                detected on subsequent solves. This includes changes to the lb, ub, 
                domain, and fixed attributes of variables. Use False only when 
                manually updating the observer with opt.update_variables() or when 
                you are certain variables are not being modified.""",
            ),
        )
        #: automatically detect changes to parameters on subsequent solves
        self.update_parameters: bool = self.declare(
            'update_parameters',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to parameter values and fixed variable values will 
                not be automatically detected on subsequent solves. Use False only 
                when manually updating the observer with 
                opt.update_parameters_and_fixed_variables() or when you are certain 
                parameters are not being modified.""",
            ),
        )
        #: automatically detect changes to named expressions on subsequent solves
        self.update_named_expressions: bool = self.declare(
            'update_named_expressions',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to Expressions will not be automatically detected on 
                subsequent solves. Use False only when manually updating the solver 
                with opt.remove_constraints() and opt.add_constraints() or when you 
                are certain Expressions are not being modified.""",
            ),
        )
        #: automatically detect changes to objectives on subsequent solves
        self.update_objectives: bool = self.declare(
            'update_objectives',
            ConfigValue(
                domain=bool,
                default=True,
                description="""
                If False, changes to objectives will not be automatically detected on 
                subsequent solves. This includes the expr and sense attributes of 
                objectives. Use False only when manually updating the solver with 
                opt.set_objective() or when you are certain objectives are not being 
                modified.""",
            ),
        )


class Reason(enum.Flag):
    no_change = 0
    bounds = 1
    fixed = 2
    domain = 4
    value = 8
    added = 16
    removed = 32
    expr = 64
    sense = 128
    sos_items = 256


class Observer(abc.ABC):
    @abc.abstractmethod
    def _update_variables(self, variables: Mapping[VarData, Reason]):
        """
        This method gets called by the ModelChangeDetector when there are
        any modifications to the set of "active" variables in the model being
        observed. By "active" variables, we mean variables
        that are used within an active component such as a constraint or
        an objective. Changes include new variables being added to the model,
        variables being removed from the model, or changes to variables
        already in the model

        Parameters
        ----------
        variables: Mapping[VarData, Reason]
            The variables and what changed about them
        """
        pass

    @abc.abstractmethod
    def _update_parameters(self, params: Mapping[ParamData, Reason]):
        """
        This method gets called by the ModelChangeDetector when there are any
        modifications to the set of "active" parameters in the model being
        observed. By "active" parameters, we mean parameters that are used within
        an active component such as a constraint or an objective. Changes include
        parameters being added to the model, parameters being removed from the model,
        or changes to parameters already in the model

        Parameters
        ----------
        params: Mapping[ParamData, Reason]
            The parameters and what changed about them
        """
        pass

    @abc.abstractmethod
    def _update_constraints(self, cons: Mapping[ConstraintData, Reason]):
        """
        This method gets called by the ModelChangeDetector when there are any
        modifications to the set of active constraints in the model being observed.
        Changes include constraints being added to the model, constraints being
        removed from the model, or changes to constraints already in the model.

        Parameters
        ----------
        cons: Mapping[ConstraintData, Reason]
            The constraints and what changed about them
        """
        pass

    @abc.abstractmethod
    def _update_sos_constraints(self, cons: Mapping[SOSConstraintData, Reason]):
        """
        This method gets called by the ModelChangeDetector when there are any
        modifications to the set of active SOS constraints in the model being
        observed. Changes include constraints being added to the model, constraints
        being removed from the model, or changes to constraints already in the model.

        Parameters
        ----------
        cons: Mapping[SOSConstraintData, Reason]
            The SOS constraints and what changed about them
        """
        pass

    @abc.abstractmethod
    def _update_objectives(self, objs: Mapping[ObjectiveData, Reason]):
        """
        This method gets called by the ModelChangeDetector when there are any
        modifications to the set of active objectives in the model being observed.
        Changes include objectives being added to the model, objectives being
        removed from the model, or changes to objectives already in the model.

        Parameters
        ----------
        objs: Mapping[ObjectiveData, Reason]
            The objectives and what changed about them
        """
        pass


def _default_reason():
    return Reason.no_change


class _Updates:
    def __init__(self, observers: Collection[Observer]) -> None:
        self.vars_to_update = DefaultComponentMap(_default_reason)
        self.params_to_update = DefaultComponentMap(_default_reason)
        self.cons_to_update = defaultdict(_default_reason)
        self.sos_to_update = defaultdict(_default_reason)
        self.objs_to_update = DefaultComponentMap(_default_reason)
        self.observers = observers

    def run(self):
        # split up new, removed, and modified variables
        new_vars = ComponentMap(
            (k, v) for k, v in self.vars_to_update.items() if v & Reason.added
        )
        other_vars = ComponentMap(
            (k, v) for k, v in self.vars_to_update.items() if not (v & Reason.added)
        )

        new_params = ComponentMap(
            (k, v) for k, v in self.params_to_update.items() if v & Reason.added
        )
        other_params = ComponentMap(
            (k, v) for k, v in self.params_to_update.items() if not (v & Reason.added)
        )

        for obs in self.observers:
            if new_vars:
                obs._update_variables(new_vars)
            if new_params:
                obs._update_parameters(new_params)
            if self.cons_to_update:
                obs._update_constraints(self.cons_to_update)
            if self.sos_to_update:
                obs._update_sos_constraints(self.sos_to_update)
            if self.objs_to_update:
                obs._update_objectives(self.objs_to_update)
            if other_vars:
                obs._update_variables(other_vars)
            if other_params:
                obs._update_parameters(other_params)

        self.clear()

    def clear(self):
        self.vars_to_update.clear()
        self.params_to_update.clear()
        self.cons_to_update.clear()
        self.sos_to_update.clear()
        self.objs_to_update.clear()


"""
There are three stages:
- identification of differences between the model and the internal data structures of the Change Detector
- synchronization of the model with the internal data structures of the ChangeDetector
- notification of the observers

The first two really happen at the same time

Update order when notifying the observers:
  - add new variables
  - add new constraints
  - add new objectives
  - remove old constraints
  - remove old objectives
  - remove old variables
  - update modified constraints
  - update modified objectives
  - update modified variables
"""


class ModelChangeDetector:
    """
    This class "watches" a pyomo model and notifies the observers when any
    changes to the model are made (but only when ModelChangeDetector.update
    is called). An example use case is for the persistent solver interfaces.

    The ModelChangeDetector considers the model to be defined by its set of
    active components and any components used by those active components. For
    example, the observers will not be notified of the addition of a variable
    if that variable is not used in any constraints.

    The Observer/ModelChangeDetector are most useful when a small number
    of changes are being made relative to the size of the model. For example,
    the persistent solver interfaces can be very efficient when repeatedly
    solving the same model but with different values for mutable parameters.

    If you know that certain changes will not be made to the model, the
    config can be modified to improve performance. For example, if you
    know that no constraints will be added to or removed from the model,
    then ``check_for_new_or_removed_constraints`` can be set to ``False``,
    which will save some time when ``update`` is called.

    Here are some usage examples:

    >>> import pyomo.environ as pyo
    >>> from typing import Mapping
    >>> from pyomo.contrib.observer.model_observer import (
    ...     AutoUpdateConfig,
    ...     Observer,
    ...     ModelChangeDetector,
    ...     Reason,
    ... )
    >>> from pyomo.core.base import (
    ...     VarData,
    ...     ParamData,
    ...     ConstraintData,
    ...     SOSConstraintData,
    ...     ObjectiveData,
    ... )
    >>> class PrintObserver(Observer):
    ...     def _update_variables(self, vars: Mapping[VarData, Reason]):
    ...         for v, r in vars.items():
    ...             print(f'{v}: {r.name}')
    ...     def _update_parameters(self, params: Mapping[ParamData, Reason]):
    ...         for p, r in params.items():
    ...             print(f'{p}: {r.name}')
    ...     def _update_constraints(self, cons: Mapping[ConstraintData, Reason]):
    ...         for c, r in cons.items():
    ...             print(f'{c}: {r.name}')
    ...     def _update_sos_constraints(self, cons: Mapping[SOSConstraintData, Reason]):
    ...         for c, r in cons.items():
    ...             print(f'{c}: {r.name}')
    ...     def _update_objectives(self, objs: Mapping[ObjectiveData, Reason]):
    ...         for o, r in objs.items():
    ...             print(f'{o}: {r.name}')
    >>> m = pyo.ConcreteModel()
    >>> obs = PrintObserver()
    >>> detector = ModelChangeDetector(m, [obs])
    >>> m.x = pyo.Var()
    >>> m.y = pyo.Var()
    >>> detector.update()  # no output because the variables are not used
    >>> m.obj = pyo.Objective(expr=m.x**2 + m.y**2)
    >>> detector.update()
    x: added
    y: added
    obj: added
    >>> del m.obj
    >>> detector.update()
    obj: removed
    x: removed
    y: removed
    >>> m.px = pyo.Param(mutable=True, initialize=1)
    >>> m.py = pyo.Param(mutable=True, initialize=1)
    >>> m.obj = pyo.Objective(expr=m.px*m.x + m.py*m.y)
    >>> detector.update()
    x: added
    y: added
    px: added
    py: added
    obj: added
    >>> m.px.value = 2
    >>> detector.update()
    px: value
    >>> detector.config.check_for_new_or_removed_constraints = False
    >>> detector.config.check_for_new_or_removed_objectives = False
    >>> detector.config.update_constraints = False
    >>> detector.config.update_vars = False
    >>> detector.config.update_parameters = True
    >>> detector.config.update_named_expressions = False
    >>> detector.config.update_objectives = False
    >>> for i in range(10):
    ...     m.py.value = i
    ...     detector.update()  # this will be faster because it is only checking for changes to parameters
    py: value
    py: value
    py: value
    py: value
    py: value
    py: value
    py: value
    py: value
    py: value
    py: value
    >>> m.c = pyo.Constraint(expr=m.y >= pyo.exp(m.x))
    >>> detector.update()  # no output because we did not check for new constraints
    >>> detector.config.check_for_new_or_removed_constraints = True
    >>> detector.update()
    c: added

    """

    def __init__(self, model: BlockData, observers: Sequence[Observer], **kwds):
        """
        Parameters
        ----------
        model: BlockData
            The model for which changes should be detected
        observers: Sequence[Observer]
            The objects to notify when changes are made to the model
        """
        self._known_active_ctypes = {Constraint, SOSConstraint, Objective, Block}
        self._observers: List[Observer] = list(observers)

        self._active_constraints: MutableMapping[
            ConstraintData, Union[RelationalExpression, None]
        ] = {}

        self._active_sos = {}

        # maps var to (lb, ub, fixed, domain, value)
        self._vars: MutableMapping[VarData, Tuple] = ComponentMap()

        # maps param to value
        self._params: MutableMapping[ParamData, float] = ComponentMap()

        self._objectives: MutableMapping[
            ObjectiveData, Tuple[Union[NumericValue, float, int, None], ObjectiveSense]
        ] = ComponentMap()  # maps objective to (expression, sense)

        # maps constraints/objectives to list of tuples (named_expr, named_expr.expr)
        self._named_expressions: MutableMapping[
            ConstraintData,
            List[Tuple[ExpressionData, Union[NumericValue, float, int, None]]],
        ] = {}
        self._obj_named_expressions: MutableMapping[
            ObjectiveData,
            List[Tuple[ExpressionData, Union[NumericValue, float, int, None]]],
        ] = ComponentMap()

        self._external_functions = ComponentMap()

        self._referenced_variables: MutableMapping[
            VarData,
            Tuple[
                MutableSet[ConstraintData],
                MutableSet[SOSConstraintData],
                MutableSet[ObjectiveData],
            ],
        ] = ComponentMap()

        self._referenced_params: MutableMapping[
            ParamData,
            Tuple[
                MutableSet[ConstraintData],
                MutableSet[SOSConstraintData],
                MutableSet[ObjectiveData],
                MutableSet[VarData],
            ],
        ] = ComponentMap()

        self._vars_referenced_by_con: MutableMapping[
            Union[ConstraintData, SOSConstraintData], MutableSet[VarData]
        ] = {}
        self._vars_referenced_by_obj: MutableMapping[
            ObjectiveData, MutableSet[VarData]
        ] = ComponentMap()
        self._params_referenced_by_con: MutableMapping[
            Union[ConstraintData, SOSConstraintData], MutableSet[ParamData]
        ] = {}
        # for when parameters show up in variable bounds
        self._params_referenced_by_var: MutableMapping[
            VarData, MutableSet[ParamData]
        ] = ComponentMap()
        self._params_referenced_by_obj: MutableMapping[
            ObjectiveData, MutableSet[ParamData]
        ] = ComponentMap()

        self.config: AutoUpdateConfig = AutoUpdateConfig()(
            value=kwds, preserve_implicit=True
        )

        self._updates = _Updates(self._observers)

        self._model: BlockData = model
        self._set_instance()

    def _add_variables(self, variables: Collection[VarData]):
        for v in variables:
            if v in self._referenced_variables:
                raise ValueError(f'Variable {v.name} has already been added')
            self._updates.vars_to_update[v] |= Reason.added
            self._referenced_variables[v] = (OrderedSet(), OrderedSet(), ComponentSet())
            self._vars[v] = (v._lb, v._ub, v.fixed, v.domain.get_interval(), v.value)
            ref_params = ComponentSet()
            for bnd in (v._lb, v._ub):
                if bnd is None or type(bnd) in native_numeric_types:
                    continue
                named_exprs, _vars, parameters, external_functions = (
                    collect_components_from_expr(bnd)
                )
                if _vars:
                    raise NotImplementedError(
                        'ModelChangeDetector does not support variables in the bounds of other variables'
                    )
                if named_exprs:
                    raise NotImplementedError(
                        'ModelChangeDetector does not support Expressions in the bounds of other variables'
                    )
                if external_functions:
                    raise NotImplementedError(
                        'ModelChangeDetector does not support external functions in the bounds of other variables'
                    )
                ref_params.update(parameters)
            self._params_referenced_by_var[v] = ref_params
            if ref_params:
                self._check_for_new_params(ref_params)
                for p in ref_params:
                    self._referenced_params[p][3].add(v)

    def _add_parameters(self, params: Collection[ParamData]):
        for p in params:
            if p in self._referenced_params:
                raise ValueError(f'Parameter {p.name} has already been added')
            self._updates.params_to_update[p] |= Reason.added
            self._referenced_params[p] = (
                OrderedSet(),
                OrderedSet(),
                ComponentSet(),
                ComponentSet(),
            )
            self._params[p] = p.value

    def _check_for_new_vars(self, variables: Collection[VarData]):
        new_vars = ComponentSet(
            v for v in variables if v not in self._referenced_variables
        )
        self._add_variables(new_vars)

    def _check_to_remove_vars(self, variables: Collection[VarData]):
        vars_to_remove = ComponentSet()
        for v in variables:
            if not any(self._referenced_variables[v]):
                vars_to_remove.add(v)
        self._remove_variables(vars_to_remove)

    def _check_for_new_params(self, params: Collection[ParamData]):
        new_params = ComponentSet(p for p in params if p not in self._referenced_params)
        self._add_parameters(new_params)

    def _check_to_remove_params(self, params: Collection[ParamData]):
        params_to_remove = ComponentSet()
        for p in params:
            if not any(self._referenced_params[p]):
                params_to_remove.add(p)
        self._remove_parameters(params_to_remove)

    def _add_constraints(self, cons: Collection[ConstraintData]):
        for con in cons:
            if con in self._active_constraints:
                raise ValueError(f'Constraint {con.name} has already been added')
            self._updates.cons_to_update[con] |= Reason.added
            self._active_constraints[con] = con.expr
            named_exprs, variables, parameters, external_functions = (
                collect_components_from_expr(con.expr)
            )
            self._check_for_new_vars(variables)
            self._check_for_new_params(parameters)
            if named_exprs:
                self._named_expressions[con] = [(e, e.expr) for e in named_exprs]
            if external_functions:
                self._external_functions[con] = external_functions
            self._vars_referenced_by_con[con] = variables
            self._params_referenced_by_con[con] = parameters
            for v in variables:
                self._referenced_variables[v][0].add(con)
            for p in parameters:
                self._referenced_params[p][0].add(con)

    def add_constraints(self, cons: Collection[ConstraintData]):
        self._add_constraints(cons)
        self._updates.run()

    def _add_sos_constraints(self, cons: Collection[SOSConstraintData]):
        for con in cons:
            if con in self._active_sos:
                raise ValueError(f'Constraint {con.name} has already been added')
            self._updates.sos_to_update[con] |= Reason.added
            sos_items = list(con.get_items())
            self._active_sos[con] = (
                [i[0] for i in sos_items],
                [i[1] for i in sos_items],
            )
            variables = ComponentSet()
            params = ComponentSet()
            for v, p in sos_items:
                variables.add(v)
                if type(p) in native_numeric_types:
                    continue
                if p.is_parameter_type():
                    params.add(p)
            self._check_for_new_vars(variables)
            self._check_for_new_params(params)
            self._vars_referenced_by_con[con] = variables
            self._params_referenced_by_con[con] = params
            for v in variables:
                self._referenced_variables[v][1].add(con)
            for p in params:
                self._referenced_params[p][1].add(con)

    def add_sos_constraints(self, cons: Collection[SOSConstraintData]):
        self._add_sos_constraints(cons)
        self._updates.run()

    def _add_objectives(self, objs: Collection[ObjectiveData]):
        for obj in objs:
            self._updates.objs_to_update[obj] |= Reason.added
            self._objectives[obj] = (obj.expr, obj.sense)
            named_exprs, variables, parameters, external_functions = (
                collect_components_from_expr(obj.expr)
            )
            self._check_for_new_vars(variables)
            self._check_for_new_params(parameters)
            if named_exprs:
                self._obj_named_expressions[obj] = [(e, e.expr) for e in named_exprs]
            if external_functions:
                self._external_functions[obj] = external_functions
            self._vars_referenced_by_obj[obj] = variables
            self._params_referenced_by_obj[obj] = parameters
            for v in variables:
                self._referenced_variables[v][2].add(obj)
            for p in parameters:
                self._referenced_params[p][2].add(obj)

    def add_objectives(self, objs: Collection[ObjectiveData]):
        self._add_objectives(objs)
        self._updates.run()

    def _remove_objectives(self, objs: Collection[ObjectiveData]):
        for obj in objs:
            if obj not in self._objectives:
                raise ValueError(
                    f'cannot remove objective {obj.name} - it was not added'
                )
            self._updates.objs_to_update[obj] |= Reason.removed
            for v in self._vars_referenced_by_obj[obj]:
                self._referenced_variables[v][2].remove(obj)
            for p in self._params_referenced_by_obj[obj]:
                self._referenced_params[p][2].remove(obj)
            self._check_to_remove_vars(self._vars_referenced_by_obj[obj])
            self._check_to_remove_params(self._params_referenced_by_obj[obj])
            del self._objectives[obj]
            self._obj_named_expressions.pop(obj, None)
            self._external_functions.pop(obj, None)
            self._vars_referenced_by_obj.pop(obj)
            self._params_referenced_by_obj.pop(obj)

    def remove_objectives(self, objs: Collection[ObjectiveData]):
        self._remove_objectives(objs)
        self._updates.run()

    def _check_for_unknown_active_components(self):
        for ctype in self._model.collect_ctypes(active=True, descend_into=True):
            if not issubclass(ctype, ActiveComponent):
                continue
            if ctype in self._known_active_ctypes:
                continue
            if ctype is Suffix:
                warnings.warn('ModelChangeDetector does not detect changes to suffixes')
                continue
            raise NotImplementedError(
                f'ModelChangeDetector does not know how to '
                f'handle components with ctype {ctype}'
            )

    def _set_instance(self):

        with PauseGC() as pgc:
            self._check_for_unknown_active_components()

            self.add_constraints(
                list(
                    self._model.component_data_objects(
                        Constraint, descend_into=True, active=True
                    )
                )
            )
            self.add_sos_constraints(
                list(
                    self._model.component_data_objects(
                        SOSConstraint, descend_into=True, active=True
                    )
                )
            )
            self.add_objectives(
                list(
                    self._model.component_data_objects(
                        Objective, descend_into=True, active=True
                    )
                )
            )

    def _remove_constraints(self, cons: Collection[ConstraintData]):
        for con in cons:
            if con not in self._active_constraints:
                raise ValueError(
                    f'Cannot remove constraint {con.name} - it was not added'
                )
            self._updates.cons_to_update[con] |= Reason.removed
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[v][0].remove(con)
            for p in self._params_referenced_by_con[con]:
                self._referenced_params[p][0].remove(con)
            self._check_to_remove_vars(self._vars_referenced_by_con[con])
            self._check_to_remove_params(self._params_referenced_by_con[con])
            self._active_constraints.pop(con)
            self._named_expressions.pop(con, None)
            self._external_functions.pop(con, None)
            self._vars_referenced_by_con.pop(con)
            self._params_referenced_by_con.pop(con)

    def remove_constraints(self, cons: Collection[ConstraintData]):
        self._remove_constraints(cons)
        self._updates.run()

    def _remove_sos_constraints(self, cons: Collection[SOSConstraintData]):
        for con in cons:
            if con not in self._active_sos:
                raise ValueError(
                    f'Cannot remove constraint {con.name} - it was not added'
                )
            self._updates.sos_to_update[con] |= Reason.removed
            for v in self._vars_referenced_by_con[con]:
                self._referenced_variables[v][1].remove(con)
            for p in self._params_referenced_by_con[con]:
                self._referenced_params[p][1].remove(con)
            self._check_to_remove_vars(self._vars_referenced_by_con[con])
            self._check_to_remove_params(self._params_referenced_by_con[con])
            self._active_sos.pop(con)
            self._vars_referenced_by_con.pop(con)
            self._params_referenced_by_con.pop(con)

    def remove_sos_constraints(self, cons: Collection[SOSConstraintData]):
        self._remove_sos_constraints(cons)
        self._updates.run()

    def _remove_variables(self, variables: Collection[VarData]):
        for v in variables:
            if v not in self._referenced_variables:
                raise ValueError(
                    f'Cannot remove variable {v.name} - it has not been added'
                )
            self._updates.vars_to_update[v] |= Reason.removed
            for p in self._params_referenced_by_var[v]:
                self._referenced_params[p][3].remove(v)
            self._check_to_remove_params(self._params_referenced_by_var[v])
            self._params_referenced_by_var.pop(v)
            if any(self._referenced_variables[v]):
                raise ValueError(
                    f'Cannot remove variable {v.name} - it is still being used by constraints/objectives'
                )
            self._referenced_variables.pop(v)
            self._vars.pop(v)

    def _remove_parameters(self, params: Collection[ParamData]):
        for p in params:
            if p not in self._referenced_params:
                raise ValueError(
                    f'Cannot remove parameter {p.name} - it has not been added'
                )
            self._updates.params_to_update[p] |= Reason.removed
            if any(self._referenced_params[p]):
                raise ValueError(
                    f'Cannot remove parameter {p.name} - it is still being used by constraints/objectives'
                )
            self._referenced_params.pop(p)
            self._params.pop(p)

    def _update_var_bounds(self, v: VarData):
        ref_params = ComponentSet()
        for bnd in (v._lb, v._ub):
            if bnd is None or type(bnd) in native_numeric_types:
                continue
            named_exprs, _vars, parameters, external_functions = (
                collect_components_from_expr(bnd)
            )
            if _vars:
                raise NotImplementedError(
                    'ModelChangeDetector does not support variables in the bounds of other variables'
                )
            if named_exprs:
                raise NotImplementedError(
                    'ModelChangeDetector does not support Expressions in the bounds of other variables'
                )
            if external_functions:
                raise NotImplementedError(
                    'ModelChangeDetector does not support external functions in the bounds of other variables'
                )
            ref_params.update(parameters)

        _ref_params = self._params_referenced_by_var[v]
        new_params = ref_params - _ref_params
        old_params = _ref_params - ref_params

        self._params_referenced_by_var[v] = ref_params

        if new_params:
            self._check_for_new_params(new_params)

        for p in new_params:
            self._referenced_params[p][3].add(v)
        for p in old_params:
            self._referenced_params[p][3].remove(v)

        if old_params:
            self._check_to_remove_params(old_params)

    def _update_variables(self, variables: Optional[Collection[VarData]] = None):
        if variables is None:
            variables = self._vars
        for v in variables:
            _lb, _ub, _fixed, _domain_interval, _value = self._vars[v]
            lb, ub, fixed, domain_interval, value = (
                v._lb,
                v._ub,
                v.fixed,
                v.domain.get_interval(),
                v.value,
            )
            reason = Reason.no_change
            if _fixed != fixed:
                reason |= Reason.fixed
            elif (_fixed or fixed) and (value != _value):
                reason |= Reason.value
            if lb is not _lb or ub is not _ub:
                reason |= Reason.bounds
            if _domain_interval != domain_interval:
                reason |= Reason.domain
            if reason:
                self._updates.vars_to_update[v] |= reason
                self._vars[v] = (lb, ub, fixed, domain_interval, value)
                if reason & Reason.bounds:
                    self._update_var_bounds(v)

    def update_variables(self, variables: Optional[Collection[VarData]] = None):
        self._update_variables(variables)
        self._updates.run()

    def _update_parameters(self, params: Optional[Collection[ParamData]] = None):
        if params is None:
            params = self._params
        for p in params:
            _val = self._params[p]
            val = p.value
            reason = Reason.no_change
            if _val != val:
                reason |= Reason.value
            if reason:
                self._updates.params_to_update[p] |= reason
                self._params[p] = val

    def update_parameters(self, params: Optional[Collection[ParamData]]):
        self._update_parameters(params)
        self._updates.run()

    def _update_con(self, con: ConstraintData):
        self._active_constraints[con] = con.expr
        named_exprs, variables, parameters, external_functions = (
            collect_components_from_expr(con.expr)
        )
        if named_exprs:
            self._named_expressions[con] = [(e, e.expr) for e in named_exprs]
        else:
            self._named_expressions.pop(con, None)
        if external_functions:
            self._external_functions[con] = external_functions
        else:
            self._external_functions.pop(con, None)

        _variables = self._vars_referenced_by_con[con]
        _parameters = self._params_referenced_by_con[con]
        new_vars = variables - _variables
        old_vars = _variables - variables
        new_params = parameters - _parameters
        old_params = _parameters - parameters

        self._vars_referenced_by_con[con] = variables
        self._params_referenced_by_con[con] = parameters

        if new_vars:
            self._check_for_new_vars(new_vars)
        if new_params:
            self._check_for_new_params(new_params)

        for v in new_vars:
            self._referenced_variables[v][0].add(con)
        for v in old_vars:
            self._referenced_variables[v][0].remove(con)
        for p in new_params:
            self._referenced_params[p][0].add(con)
        for p in old_params:
            self._referenced_params[p][0].remove(con)

        if old_vars:
            self._check_to_remove_vars(old_vars)
        if old_params:
            self._check_to_remove_params(old_params)

    def _update_constraints(self, cons: Optional[Collection[ConstraintData]] = None):
        if cons is None:
            cons = self._active_constraints
        for c in cons:
            reason = Reason.no_change
            if c.expr is not self._active_constraints[c]:
                reason |= Reason.expr
            if reason:
                self._updates.cons_to_update[c] |= reason
                self._update_con(c)

    def update_constraints(self, cons: Optional[Collection[ConstraintData]] = None):
        self._update_constraints(cons)
        self._updates.run()

    def _update_sos_con(self, con: SOSConstraintData):
        sos_items = list(con.get_items())
        self._active_sos[con] = ([i[0] for i in sos_items], [i[1] for i in sos_items])
        variables = ComponentSet()
        parameters = ComponentSet()
        for v, p in sos_items:
            variables.add(v)
            if type(p) in native_numeric_types:
                continue
            if p.is_parameter_type():
                parameters.add(p)

        _variables = self._vars_referenced_by_con[con]
        _parameters = self._params_referenced_by_con[con]
        new_vars = variables - _variables
        old_vars = _variables - variables
        new_params = parameters - _parameters
        old_params = _parameters - parameters

        self._vars_referenced_by_con[con] = variables
        self._params_referenced_by_con[con] = parameters

        if new_vars:
            self._check_for_new_vars(new_vars)
        if new_params:
            self._check_for_new_params(new_params)

        for v in new_vars:
            self._referenced_variables[v][1].add(con)
        for v in old_vars:
            self._referenced_variables[v][1].remove(con)
        for p in new_params:
            self._referenced_params[p][1].add(con)
        for p in old_params:
            self._referenced_params[p][1].remove(con)

        if old_vars:
            self._check_to_remove_vars(old_vars)
        if old_params:
            self._check_to_remove_params(old_params)

    def _update_sos_constraints(
        self, cons: Optional[Collection[SOSConstraintData]] = None
    ):
        if cons is None:
            cons = self._active_sos
        for c in cons:
            reason = Reason.no_change
            _vlist, _plist = self._active_sos[c]
            sos_items = list(c.get_items())
            vlist = [i[0] for i in sos_items]
            plist = [i[1] for i in sos_items]
            needs_update = False
            if len(_vlist) != len(vlist) or len(_plist) != len(plist):
                needs_update = True
            else:
                for v1, v2 in zip(_vlist, vlist):
                    if v1 is not v2:
                        needs_update = True
                        break
                for p1, p2 in zip(_plist, plist):
                    if p1 is not p2:
                        needs_update = True
                        break
            if needs_update:
                reason |= Reason.sos_items
                self._updates.sos_to_update[c] |= reason
                self._update_sos_con(c)

    def update_sos_constraints(
        self, cons: Optional[Collection[SOSConstraintData]] = None
    ):
        self._update_sos_constraints(cons)
        self._updates.run()

    def _update_obj_expr(self, obj: ObjectiveData):
        named_exprs, variables, parameters, external_functions = (
            collect_components_from_expr(obj.expr)
        )
        if named_exprs:
            self._obj_named_expressions[obj] = [(e, e.expr) for e in named_exprs]
        else:
            self._obj_named_expressions.pop(obj, None)
        if external_functions:
            self._external_functions[obj] = external_functions
        else:
            self._external_functions.pop(obj, None)

        _variables = self._vars_referenced_by_obj[obj]
        _parameters = self._params_referenced_by_obj[obj]
        new_vars = variables - _variables
        old_vars = _variables - variables
        new_params = parameters - _parameters
        old_params = _parameters - parameters

        self._vars_referenced_by_obj[obj] = variables
        self._params_referenced_by_obj[obj] = parameters

        if new_vars:
            self._check_for_new_vars(new_vars)
        if new_params:
            self._check_for_new_params(new_params)

        for v in new_vars:
            self._referenced_variables[v][2].add(obj)
        for v in old_vars:
            self._referenced_variables[v][2].remove(obj)
        for p in new_params:
            self._referenced_params[p][2].add(obj)
        for p in old_params:
            self._referenced_params[p][2].remove(obj)

        if old_vars:
            self._check_to_remove_vars(old_vars)
        if old_params:
            self._check_to_remove_params(old_params)

    def _update_objectives(self, objs: Optional[Collection[ObjectiveData]] = None):
        if objs is None:
            objs = self._objectives
        for obj in objs:
            reason = Reason.no_change
            _expr, _sense = self._objectives[obj]
            if _expr is not obj.expr:
                reason |= Reason.expr
            if _sense != obj.sense:
                reason |= Reason.sense
            if reason:
                self._updates.objs_to_update[obj] |= reason
                self._objectives[obj] = (obj.expr, obj.sense)
                if reason & Reason.expr:
                    self._update_obj_expr(obj)

    def update_objectives(self, objs: Optional[Collection[ObjectiveData]] = None):
        self._update_objectives(objs)
        self._updates.run()

    def _check_for_new_or_removed_sos(self):
        new_sos = []
        old_sos = []
        current_sos_set = OrderedSet(
            self._model.component_data_objects(
                SOSConstraint, descend_into=True, active=True
            )
        )
        for c in current_sos_set:
            if c not in self._active_sos:
                new_sos.append(c)
        for c in self._active_sos:
            if c not in current_sos_set:
                old_sos.append(c)
        return new_sos, old_sos

    def _check_for_new_or_removed_constraints(self):
        new_cons = []
        old_cons = []
        current_cons_set = OrderedSet(
            self._model.component_data_objects(
                Constraint, descend_into=True, active=True
            )
        )
        for c in current_cons_set:
            if c not in self._active_constraints:
                new_cons.append(c)
        for c in self._active_constraints:
            if c not in current_cons_set:
                old_cons.append(c)
        return new_cons, old_cons

    def _check_for_named_expression_changes(self):
        for con, ne_list in self._named_expressions.items():
            for named_expr, old_expr in ne_list:
                if named_expr.expr is not old_expr:
                    self._updates.cons_to_update[con] |= Reason.expr
                    self._update_con(con)
                    break
        for obj, ne_list in self._obj_named_expressions.items():
            for named_expr, old_expr in ne_list:
                if named_expr.expr is not old_expr:
                    self._updates.objs_to_update[obj] |= Reason.expr
                    self._update_obj_expr(obj)
                    break

    def _check_for_new_or_removed_objectives(self):
        new_objs = []
        old_objs = []
        current_objs_set = ComponentSet(
            self._model.component_data_objects(
                Objective, descend_into=True, active=True
            )
        )
        for obj in current_objs_set:
            if obj not in self._objectives:
                new_objs.append(obj)
        for obj in self._objectives.keys():
            if obj not in current_objs_set:
                old_objs.append(obj)
        return new_objs, old_objs

    def update(self, timer: Optional[HierarchicalTimer] = None, **kwds):
        """
        Check for changes to the model and notify the observers.

        Parameters
        ----------
        timer: Optional[HierarchicalTimer]
            The timer to use for tracking how much time is spent detecting
            different kinds of changes
        """

        """
        When possible, it is better to add new constraints before removing old 
        constraints. This prevents unnecessarily removing and adding variables.
        If a constraint is removed, any variables that are used only by that 
        constraint will be removed. If there is a new constraint that uses 
        the same variable, then we don't actually need to remove the variable.
        This is hard to avoid when we are modifying a constraint or changing
        the objective. When the objective changes, we remove the old one 
        first just because most things don't handle multiple objectives.

        We check for changes to constraints/objectives before variables/parameters 
        so that we don't waste time updating a variable/parameter that is going to 
        get removed.
        """
        if timer is None:
            timer = HierarchicalTimer()
        config: AutoUpdateConfig = self.config(value=kwds, preserve_implicit=True)

        with PauseGC() as pgc:
            self._check_for_unknown_active_components()

            if config.check_for_new_or_removed_constraints:
                new_cons, old_cons = self._check_for_new_or_removed_constraints()
                new_sos, old_sos = self._check_for_new_or_removed_sos()
            else:
                new_cons = []
                old_cons = []
                new_sos = []
                old_sos = []

            if config.check_for_new_or_removed_objectives:
                new_objs, old_objs = self._check_for_new_or_removed_objectives()
            else:
                new_objs = []
                old_objs = []

            if new_cons:
                self._add_constraints(new_cons)
            if new_sos:
                self._add_sos_constraints(new_sos)
            if new_objs:
                self._add_objectives(new_objs)

            if old_cons:
                self._remove_constraints(old_cons)
            if old_sos:
                self._remove_sos_constraints(old_sos)
            if old_objs:
                self._remove_objectives(old_objs)

            if config.update_constraints:
                self._update_constraints()
                self._update_sos_constraints()
            if config.update_objectives:
                self._update_objectives()

            if config.update_named_expressions:
                self._check_for_named_expression_changes()

            if config.update_vars:
                self._update_variables()

            if config.update_parameters:
                self._update_parameters()

            self._updates.run()

    def get_variables_impacted_by_param(self, p: ParamData):
        return list(self._referenced_params[p][3])

    def get_constraints_impacted_by_param(self, p: ParamData):
        return list(self._referenced_params[p][0])

    def get_constraints_impacted_by_var(self, v: VarData):
        return list(self._referenced_variables[v][0])

    def get_objectives_impacted_by_param(self, p: ParamData):
        return list(self._referenced_params[p][2])

    def get_objectives_impacted_by_var(self, v: VarData):
        return list(self._referenced_variables[v][2])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Unit Tests for interfacing with sIPOPT and k_aug
"""

from io import StringIO
import logging
import pyomo.common.unittest as unittest

from pyomo.environ import ConcreteModel, Param, Var, Block, Suffix, value
from pyomo.opt import SolverFactory
from pyomo.dae import ContinuousSet
from pyomo.common.dependencies import scipy_available
from pyomo.common.log import LoggingIntercept
from pyomo.common.collections import ComponentMap
from pyomo.core.expr import identify_variables
from pyomo.contrib.sensitivity_toolbox.sens import sipopt, kaug, sensitivity_calculation
import pyomo.contrib.sensitivity_toolbox.examples.parameter as param_ex
import pyomo.contrib.sensitivity_toolbox.examples.parameter_kaug as param_kaug_ex
import pyomo.contrib.sensitivity_toolbox.examples.feedbackController as fc
import pyomo.contrib.sensitivity_toolbox.examples.rangeInequality as ri
import pyomo.contrib.sensitivity_toolbox.examples.HIV_Transmission as hiv

opt = SolverFactory('ipopt_sens', solver_io='nl')
opt_kaug = SolverFactory('k_aug', solver_io='nl')
opt_dotsens = SolverFactory('dot_sens', solver_io='nl')


class FunctionDeprecationTest(unittest.TestCase):
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_sipopt_deprecated(self):
        m = param_ex.create_model()
        m.perturbed_eta1 = Param(initialize=4.0)
        m.perturbed_eta2 = Param(initialize=1.0)

        output = StringIO()
        with LoggingIntercept(
            output, 'pyomo.contrib.sensitivity_toolbox', logging.WARNING
        ):
            sipopt(
                m,
                [m.eta1, m.eta1],
                [m.perturbed_eta1, m.perturbed_eta2],
                cloneModel=False,
            )
        self.assertIn(
            "DEPRECATED: The sipopt function has been deprecated. Use the "
            "sensitivity_calculation() function with method='sipopt' to access",
            output.getvalue().replace('\n', ' '),
        )

    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_sipopt_equivalent(self):
        m1 = param_ex.create_model()
        m1.perturbed_eta1 = Param(initialize=4.0)
        m1.perturbed_eta2 = Param(initialize=1.0)

        m2 = param_ex.create_model()
        m2.perturbed_eta1 = Param(initialize=4.0)
        m2.perturbed_eta2 = Param(initialize=1.0)

        m11 = sipopt(
            m1,
            [m1.eta1, m1.eta2],
            [m1.perturbed_eta1, m1.perturbed_eta2],
            cloneModel=True,
        )
        m22 = sensitivity_calculation(
            'sipopt',
            m2,
            [m2.eta1, m2.eta2],
            [m2.perturbed_eta1, m2.perturbed_eta2],
            cloneModel=True,
        )
        out1 = StringIO()
        out2 = StringIO()
        m11._SENSITIVITY_TOOLBOX_DATA.constList.pprint(ostream=out1)
        m22._SENSITIVITY_TOOLBOX_DATA.constList.pprint(ostream=out2)
        self.assertMultiLineEqual(out1.getvalue(), out2.getvalue())

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_kaug_deprecated(self):
        m = param_ex.create_model()
        m.perturbed_eta1 = Param(initialize=4.0)
        m.perturbed_eta2 = Param(initialize=1.0)

        output = StringIO()
        with LoggingIntercept(
            output, 'pyomo.contrib.sensitivity_toolbox', logging.WARNING
        ):
            kaug(
                m,
                [m.eta1, m.eta1],
                [m.perturbed_eta1, m.perturbed_eta2],
                cloneModel=False,
            )
        self.assertIn(
            "DEPRECATED: The kaug function has been deprecated. Use the "
            "sensitivity_calculation() function with method='k_aug'",
            output.getvalue().replace('\n', ' '),
        )

    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_kaug_equivalent(self):
        m1 = param_ex.create_model()
        m1.perturbed_eta1 = Param(initialize=4.0)
        m1.perturbed_eta2 = Param(initialize=1.0)

        m2 = param_ex.create_model()
        m2.perturbed_eta1 = Param(initialize=4.0)
        m2.perturbed_eta2 = Param(initialize=1.0)

        m11 = kaug(
            m1,
            [m1.eta1, m1.eta2],
            [m1.perturbed_eta1, m1.perturbed_eta2],
            cloneModel=True,
        )
        m22 = sensitivity_calculation(
            'k_aug',
            m2,
            [m2.eta1, m2.eta2],
            [m2.perturbed_eta1, m2.perturbed_eta2],
            cloneModel=True,
        )
        out1 = StringIO()
        out2 = StringIO()
        m11._SENSITIVITY_TOOLBOX_DATA.constList.pprint(ostream=out1)
        m22._SENSITIVITY_TOOLBOX_DATA.constList.pprint(ostream=out2)
        self.assertMultiLineEqual(out1.getvalue(), out2.getvalue())


class TestSensitivityToolbox(unittest.TestCase):
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_bad_arg(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 1))

        m.a = Param(initialize=1, mutable=True)
        m.b = Param(initialize=2, mutable=True)
        m.c = Param(initialize=3, mutable=False)

        m.x = Var(m.t)

        list_one = [m.a, m.b]
        list_two = [m.a, m.b, m.c]
        list_three = [m.a, m.x]
        list_four = [m.a, m.c]

        # verify ValueError thrown when param and perturb list are different
        # lengths
        msg = "Length of paramList argument does not equal length of perturbList"
        with self.assertRaisesRegex(ValueError, msg):
            Result = sensitivity_calculation('sipopt', m, list_one, list_two)

        # verify ValueError thrown when param list has an unmutable param
        msg = "Parameters within paramList must be mutable"
        with self.assertRaisesRegex(ValueError, msg):
            Result = sensitivity_calculation('sipopt', m, list_four, list_one)

        # verify ValueError thrown when param list has an unfixed var.
        msg = "Specified \"parameter\" variables must be fixed"
        with self.assertRaisesRegex(ValueError, msg) as context:
            Result = sensitivity_calculation('sipopt', m, list_three, list_one)

    # test feedbackController Solution when the model gets cloned
    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_clonedModel_soln(self):
        m_orig = fc.create_model()
        fc.initialize_model(m_orig, 100)

        m_orig.perturbed_a = Param(initialize=-0.25)
        m_orig.perturbed_H = Param(initialize=0.55)

        m_sipopt = sensitivity_calculation(
            'sipopt',
            m_orig,
            [m_orig.a, m_orig.H],
            [m_orig.perturbed_a, m_orig.perturbed_H],
            cloneModel=True,
        )

        # verify cloned model has _SENSITIVITY_TOOLBOX_DATA block
        # and original model is untouched
        self.assertFalse(m_sipopt == m_orig)

        self.assertTrue(
            hasattr(m_sipopt, '_SENSITIVITY_TOOLBOX_DATA')
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.ctype is Block
        )

        self.assertFalse(hasattr(m_orig, '_SENSITIVITY_TOOLBOX_DATA'))
        self.assertFalse(hasattr(m_orig, 'b'))

        # verify variable declaration
        self.assertTrue(
            hasattr(m_sipopt._SENSITIVITY_TOOLBOX_DATA, 'a')
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.a.ctype is Var
        )
        self.assertTrue(
            hasattr(m_sipopt._SENSITIVITY_TOOLBOX_DATA, 'H')
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.H.ctype is Var
        )

        # verify suffixes
        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_0')
            and m_sipopt.sens_state_0.ctype is Suffix
            and m_sipopt.sens_state_0[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_sipopt.sens_state_0[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_1')
            and m_sipopt.sens_state_1.ctype is Suffix
            and m_sipopt.sens_state_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_sipopt.sens_state_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_value_1')
            and m_sipopt.sens_state_value_1.ctype is Suffix
            and m_sipopt.sens_state_value_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H]
            == 0.55
            and m_sipopt.sens_state_value_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a]
            == -0.25
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_init_constr')
            and m_sipopt.sens_init_constr.ctype is Suffix
            and m_sipopt.sens_init_constr[
                m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[1]
            ]
            == 1
            and m_sipopt.sens_init_constr[
                m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[2]
            ]
            == 2
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1')
            and m_sipopt.sens_sol_state_1.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1[m_sipopt.F[15]], -0.00102016765, 8
        )

        # These tests require way too much precision for something that
        # just needs to enforce that bounds are not active...
        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1_z_L')
            and m_sipopt.sens_sol_state_1_z_L.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1_z_L[m_sipopt.u[15]], -2.181712e-09, 13
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1_z_U')
            and m_sipopt.sens_sol_state_1_z_U.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1_z_U[m_sipopt.u[15]], 6.580899e-09, 13
        )

        # verify deactivated constraints for cloned model
        self.assertFalse(
            m_sipopt.FDiffCon[0].active
            and m_sipopt.FDiffCon[7.5].active
            and m_sipopt.FDiffCon[15].active
        )

        self.assertFalse(
            m_sipopt.x_dot[0].active
            and m_sipopt.x_dot[7.5].active
            and m_sipopt.x_dot[15].active
        )

        # verify constraints on original model are still active
        self.assertTrue(
            m_orig.FDiffCon[0].active
            and m_orig.FDiffCon[7.5].active
            and m_orig.FDiffCon[15].active
        )

        self.assertTrue(
            m_orig.x_dot[0].active
            and m_orig.x_dot[7.5].active
            and m_orig.x_dot[15].active
        )

        # verify solution
        # NOTE: This is the solution to the original problem,
        # not the result of any sensitivity update.
        self.assertAlmostEqual(value(m_sipopt.J), 0.0048956783, 8)

    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_noClone_soln(self):
        m_orig = fc.create_model()
        fc.initialize_model(m_orig, 100)

        m_orig.perturbed_a = Param(initialize=-0.25)
        m_orig.perturbed_H = Param(initialize=0.55)

        m_sipopt = sensitivity_calculation(
            'sipopt',
            m_orig,
            [m_orig.a, m_orig.H],
            [m_orig.perturbed_a, m_orig.perturbed_H],
            cloneModel=False,
        )

        self.assertTrue(m_sipopt == m_orig)

        # test _SENSITIVITY_TOOLBOX_DATA block exists
        self.assertTrue(
            hasattr(m_orig, '_SENSITIVITY_TOOLBOX_DATA')
            and m_orig._SENSITIVITY_TOOLBOX_DATA.ctype is Block
        )

        # test variable declaration
        self.assertTrue(
            hasattr(m_sipopt._SENSITIVITY_TOOLBOX_DATA, 'a')
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.a.ctype is Var
        )
        self.assertTrue(
            hasattr(m_sipopt._SENSITIVITY_TOOLBOX_DATA, 'H')
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.H.ctype is Var
        )

        # test for suffixes
        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_0')
            and m_sipopt.sens_state_0.ctype is Suffix
            and m_sipopt.sens_state_0[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_sipopt.sens_state_0[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_1')
            and m_sipopt.sens_state_1.ctype is Suffix
            and m_sipopt.sens_state_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_sipopt.sens_state_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_state_value_1')
            and m_sipopt.sens_state_value_1.ctype is Suffix
            and m_sipopt.sens_state_value_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.H]
            == 0.55
            and m_sipopt.sens_state_value_1[m_sipopt._SENSITIVITY_TOOLBOX_DATA.a]
            == -0.25
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_init_constr')
            and m_sipopt.sens_init_constr.ctype is Suffix
            and m_sipopt.sens_init_constr[
                m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[1]
            ]
            == 1
            and m_sipopt.sens_init_constr[
                m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[2]
            ]
            == 2
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1')
            and m_sipopt.sens_sol_state_1.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1[m_sipopt.F[15]], -0.00102016765, 8
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1_z_L')
            and m_sipopt.sens_sol_state_1_z_L.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1_z_L[m_sipopt.u[15]], -2.181712e-09, 13
        )

        self.assertTrue(
            hasattr(m_sipopt, 'sens_sol_state_1_z_U')
            and m_sipopt.sens_sol_state_1_z_U.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_sipopt.sens_sol_state_1_z_U[m_sipopt.u[15]], 6.580899e-09, 13
        )

        # verify deactivated constraints on model
        self.assertFalse(
            m_sipopt.FDiffCon[0].active
            and m_sipopt.FDiffCon[7.5].active
            and m_sipopt.FDiffCon[15].active
        )

        self.assertFalse(
            m_sipopt.x_dot[0].active
            and m_sipopt.x_dot[7.5].active
            and m_sipopt.x_dot[15].active
        )

        # test model solution
        # NOTE:
        # ipopt_sens does not alter the values in the model,
        # so all this test is doing is making sure that the
        # objective value doesn't change. This test does nothing to
        # check values of the perturbed solution.
        self.assertAlmostEqual(value(m_sipopt.J), 0.0048956783, 8)

    # test indexed param mapping to var and perturbed values
    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_indexedParamsMapping(self):
        m = hiv.create_model()
        hiv.initialize_model(m, 10, 5, 1)

        m.epsDelta = Param(initialize=0.75001)

        q_del = {}
        q_del[(0, 0)] = 1.001
        q_del[(0, 1)] = 1.002
        q_del[(1, 0)] = 1.003
        q_del[(1, 1)] = 1.004
        q_del[(2, 0)] = 0.83001
        q_del[(2, 1)] = 0.83002
        q_del[(3, 0)] = 0.42001
        q_del[(4, 0)] = 0.17001
        m.qqDelta = Param(m.ij, initialize=q_del)

        m.aaDelta = Param(initialize=0.0001001)

        m_sipopt = sensitivity_calculation(
            'sipopt', m, [m.eps, m.qq, m.aa], [m.epsDelta, m.qqDelta, m.aaDelta]
        )

        # Make sure Param constraints have the correct form, i.e.
        # 0 <= _SENSITIVITY_TOOLBOX_DATA.PARAM_NAME - PARAM_NAME <= 0
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[1].lower, 0.0)
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[1].upper, 0.0)
        self.assertEqual(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[1].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.eps - eps',
        )
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[6].lower, 0.0)
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[6].upper, 0.0)
        self.assertEqual(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[6].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.qq[2,0] - qq[2,0]',
        )
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[10].lower, 0.0)
        self.assertEqual(m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[10].upper, 0.0)
        self.assertEqual(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.paramConst[10].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.aa - aa',
        )

    # test Constraint substitution
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_constraintSub(self):
        m = ri.create_model()

        m.pert_a = Param(initialize=0.01)
        m.pert_b = Param(initialize=1.01)

        m_sipopt = sensitivity_calculation(
            'sipopt', m, [m.a, m.b], [m.pert_a, m.pert_b]
        )

        # verify substitutions in equality constraint
        self.assertTrue(
            m_sipopt.C_equal.lower.ctype is Param
            and m_sipopt.C_equal.upper.ctype is Param
        )
        self.assertFalse(m_sipopt.C_equal.active)

        self.assertTrue(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[3].lower == 0.0
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[3].upper == 0.0
            and len(
                list(
                    identify_variables(
                        m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[3].body
                    )
                )
            )
            == 2
        )

        # verify substitutions in one-sided bounded constraint
        self.assertTrue(
            m_sipopt.C_singleBnd.lower is None
            and m_sipopt.C_singleBnd.upper.ctype is Param
        )
        self.assertFalse(m_sipopt.C_singleBnd.active)

        self.assertTrue(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[4].lower is None
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[4].upper == 0.0
            and len(
                list(
                    identify_variables(
                        m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[4].body
                    )
                )
            )
            == 2
        )

        # verify substitutions in ranged inequality constraint
        self.assertTrue(
            m_sipopt.C_rangedIn.lower.ctype is Param
            and m_sipopt.C_rangedIn.upper.ctype is Param
        )
        self.assertFalse(m_sipopt.C_rangedIn.active)

        self.assertTrue(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[1].lower is None
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[1].upper == 0.0
            and len(
                list(
                    identify_variables(
                        m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[1].body
                    )
                )
            )
            == 2
        )

        self.assertTrue(
            m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[2].lower is None
            and m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[2].upper == 0.0
            and len(
                list(
                    identify_variables(
                        m_sipopt._SENSITIVITY_TOOLBOX_DATA.constList[2].body
                    )
                )
            )
            == 2
        )

    # Test example `parameter.py`
    @unittest.skipIf(not opt.available(False), "ipopt_sens is not available")
    def test_parameter_example(self):
        d = param_ex.run_example()

        d_correct = {
            'eta1': 4.5,
            'eta2': 1.0,
            'x1_init': 0.15,
            'x2_init': 0.15,
            'x3_init': 0.0,
            'cost_sln': 0.5,
            'x1_sln': 0.5,
            'x2_sln': 0.5,
            'x3_sln': 0.0,
            'eta1_pert': 4.0,
            'eta2_pert': 1.0,
            'x1_pert': 0.3333333,
            'x2_pert': 0.6666667,
            'x3_pert': 0.0,
            'cost_pert': 0.55555556,
        }

        for k in d_correct.keys():
            # Check each element of the 'correct' dictionary against the returned
            # dictionary to 3 decimal places
            self.assertAlmostEqual(d[k], d_correct[k], 3)

    # Test kaug
    # Perform the same tests as for sipopt
    # test feedbackController Solution when the model gets cloned
    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_kaug_clonedModel_soln_kaug(self):
        m_orig = fc.create_model()
        fc.initialize_model(m_orig, 100)

        m_orig.perturbed_a = Param(initialize=-0.25)
        m_orig.perturbed_H = Param(initialize=0.55)

        m_kaug = sensitivity_calculation(
            'k_aug',
            m_orig,
            [m_orig.a, m_orig.H],
            [m_orig.perturbed_a, m_orig.perturbed_H],
            cloneModel=True,
        )

        ptb_map = ComponentMap()
        ptb_map[m_kaug.a] = value(-(m_orig.perturbed_a - m_orig.a))
        ptb_map[m_kaug.H] = value(-(m_orig.perturbed_H - m_orig.H))

        # verify cloned model has _SENSITIVITY_TOOLBOX_DATA block
        # and original model is untouched
        self.assertIsNot(m_kaug, m_orig)

        self.assertTrue(
            hasattr(m_kaug, '_SENSITIVITY_TOOLBOX_DATA')
            and m_kaug._SENSITIVITY_TOOLBOX_DATA.ctype is Block
        )

        self.assertFalse(hasattr(m_orig, '_SENSITIVITY_TOOLBOX_DATA'))
        self.assertFalse(hasattr(m_orig, 'b'))

        # verify variable declaration
        self.assertTrue(
            hasattr(m_kaug._SENSITIVITY_TOOLBOX_DATA, 'a')
            and m_kaug._SENSITIVITY_TOOLBOX_DATA.a.ctype is Var
        )
        self.assertTrue(
            hasattr(m_kaug._SENSITIVITY_TOOLBOX_DATA, 'H')
            and m_kaug._SENSITIVITY_TOOLBOX_DATA.H.ctype is Var
        )

        # verify suffixes
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_0')
            and m_kaug.sens_state_0.ctype is Suffix
            and m_kaug.sens_state_0[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_kaug.sens_state_0[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_1')
            and m_kaug.sens_state_1.ctype is Suffix
            and m_kaug.sens_state_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_kaug.sens_state_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_value_1')
            and m_kaug.sens_state_value_1.ctype is Suffix
            and m_kaug.sens_state_value_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 0.55
            and m_kaug.sens_state_value_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == -0.25
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_init_constr')
            and m_kaug.sens_init_constr.ctype is Suffix
            and m_kaug.sens_init_constr[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]]
            == 1
            and m_kaug.sens_init_constr[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]]
            == 2
        )
        self.assertTrue(hasattr(m_kaug, 'DeltaP'))
        self.assertTrue(m_kaug.DeltaP.ctype is Suffix)
        self.assertEqual(
            m_kaug.DeltaP[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]],
            ptb_map[m_kaug.a],
        )
        self.assertEqual(
            m_kaug.DeltaP[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]],
            ptb_map[m_kaug.H],
        )
        self.assertTrue(
            hasattr(m_kaug, 'dcdp')
            and m_kaug.dcdp.ctype is Suffix
            and m_kaug.dcdp[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]] == 1
            and m_kaug.dcdp[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]] == 2
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_sol_state_1')
            and m_kaug.sens_sol_state_1.ctype is Suffix
        )

        self.assertTrue(
            hasattr(m_kaug, 'ipopt_zL_in') and m_kaug.ipopt_zL_in.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_kaug.ipopt_zL_in[m_kaug.u[15]], 7.162686166847096e-09, 13
        )

        self.assertTrue(
            hasattr(m_kaug, 'ipopt_zU_in') and m_kaug.ipopt_zU_in.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_kaug.ipopt_zU_in[m_kaug.u[15]], -1.2439730261288605e-08, 13
        )
        # verify deactivated constraints for cloned model
        self.assertFalse(
            m_kaug.FDiffCon[0].active
            and m_kaug.FDiffCon[7.5].active
            and m_kaug.FDiffCon[15].active
        )

        self.assertFalse(
            m_kaug.x_dot[0].active
            and m_kaug.x_dot[7.5].active
            and m_kaug.x_dot[15].active
        )

        # verify constraints on original model are still active
        self.assertTrue(
            m_orig.FDiffCon[0].active
            and m_orig.FDiffCon[7.5].active
            and m_orig.FDiffCon[15].active
        )

        self.assertTrue(
            m_orig.x_dot[0].active
            and m_orig.x_dot[7.5].active
            and m_orig.x_dot[15].active
        )

        # verify solution
        # This is the only test that verifies the solution. Here we
        # verify the objective function value, which is a weak test.
        self.assertAlmostEqual(value(m_kaug.J), 0.002633263921107476, 8)
        # The original objective function value is 0.0048.
        # The answer from an attempt to reproduce this calculation with
        # PyNumero, with no inertia correction, seems to be 0.00047.
        # "Real solution" with the full nonlinear problem is 0.00138.
        # 0.00263 is the value we get after sensitivity update with k_aug
        # using MA57 and k_aug's default regularization strategy.

    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_noClone_soln_kaug(self):
        m_orig = fc.create_model()
        fc.initialize_model(m_orig, 100)

        m_orig.perturbed_a = Param(initialize=-0.25)
        m_orig.perturbed_H = Param(initialize=0.55)

        m_kaug = sensitivity_calculation(
            'k_aug',
            m_orig,
            [m_orig.a, m_orig.H],
            [m_orig.perturbed_a, m_orig.perturbed_H],
            cloneModel=False,
        )

        ptb_map = ComponentMap()
        ptb_map[m_kaug.a] = value(-(m_kaug.perturbed_a - m_kaug.a))
        ptb_map[m_kaug.H] = value(-(m_kaug.perturbed_H - m_kaug.H))

        self.assertTrue(m_kaug == m_orig)

        # verify suffixes
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_0')
            and m_kaug.sens_state_0.ctype is Suffix
            and m_kaug.sens_state_0[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_kaug.sens_state_0[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_1')
            and m_kaug.sens_state_1.ctype is Suffix
            and m_kaug.sens_state_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 2
            and m_kaug.sens_state_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == 1
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_state_value_1')
            and m_kaug.sens_state_value_1.ctype is Suffix
            and m_kaug.sens_state_value_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.H] == 0.55
            and m_kaug.sens_state_value_1[m_kaug._SENSITIVITY_TOOLBOX_DATA.a] == -0.25
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_init_constr')
            and m_kaug.sens_init_constr.ctype is Suffix
            and m_kaug.sens_init_constr[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]]
            == 1
            and m_kaug.sens_init_constr[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]]
            == 2
        )
        self.assertTrue(hasattr(m_kaug, 'DeltaP'))
        self.assertIs(m_kaug.DeltaP.ctype, Suffix)
        self.assertEqual(
            m_kaug.DeltaP[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]],
            ptb_map[m_kaug.a],
        )
        self.assertEqual(
            m_kaug.DeltaP[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]],
            ptb_map[m_kaug.H],
        )
        self.assertTrue(
            hasattr(m_kaug, 'dcdp')
            and m_kaug.dcdp.ctype is Suffix
            and m_kaug.dcdp[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1]] == 1
            and m_kaug.dcdp[m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[2]] == 2
        )
        self.assertTrue(
            hasattr(m_kaug, 'sens_sol_state_1')
            and m_kaug.sens_sol_state_1.ctype is Suffix
        )

        self.assertTrue(
            hasattr(m_kaug, 'ipopt_zL_in') and m_kaug.ipopt_zL_in.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_kaug.ipopt_zL_in[m_kaug.u[15]], 7.162686166847096e-09, 13
        )

        self.assertTrue(
            hasattr(m_kaug, 'ipopt_zU_in') and m_kaug.ipopt_zU_in.ctype is Suffix
        )
        self.assertAlmostEqual(
            m_kaug.ipopt_zU_in[m_kaug.u[15]], -1.2439730261288605e-08, 13
        )
        # verify deactivated constraints for cloned model
        self.assertFalse(
            m_kaug.FDiffCon[0].active
            and m_kaug.FDiffCon[7.5].active
            and m_kaug.FDiffCon[15].active
        )

        self.assertFalse(
            m_kaug.x_dot[0].active
            and m_kaug.x_dot[7.5].active
            and m_kaug.x_dot[15].active
        )

        # verify solution
        # This is the only test that verifies the solution. Here we
        # verify the objective function value, which is a weak test.
        self.assertAlmostEqual(value(m_kaug.J), 0.002633263921107476, 8)
        # The original objective function value is 0.0048.
        # The answer from an attempt to reproduce this calculation with
        # PyNumero, with no inertia correction, seems to be 0.00047.
        # "Real solution" with the full nonlinear problem is 0.00138.
        # 0.00263 is the value we get after sensitivity update with k_aug
        # using MA57 and k_aug's default regularization strategy.

    # test indexed param mapping to var and perturbed values
    @unittest.skipIf(not scipy_available, "scipy is required for this test")
    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_indexedParamsMapping_kaug(self):
        m = hiv.create_model()
        hiv.initialize_model(m, 10, 5, 1)

        m.epsDelta = Param(initialize=0.75001)

        q_del = {}
        q_del[(0, 0)] = 1.001
        q_del[(0, 1)] = 1.002
        q_del[(1, 0)] = 1.003
        q_del[(1, 1)] = 1.004
        q_del[(2, 0)] = 0.83001
        q_del[(2, 1)] = 0.83002
        q_del[(3, 0)] = 0.42001
        q_del[(4, 0)] = 0.17001
        m.qqDelta = Param(m.ij, initialize=q_del)

        m.aaDelta = Param(initialize=0.0001001)

        m_kaug = sensitivity_calculation(
            'k_aug', m, [m.eps, m.qq, m.aa], [m.epsDelta, m.qqDelta, m.aaDelta]
        )

        # Make sure Param constraints have the correct form, i.e.
        # 0 <= _SENSITIVITY_TOOLBOX_DATA.PARAM_NAME - PARAM_NAME <= 0
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1].lower, 0.0)
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1].upper, 0.0)
        self.assertEqual(
            m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[1].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.eps - eps',
        )
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[6].lower, 0.0)
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[6].upper, 0.0)
        self.assertEqual(
            m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[6].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.qq[2,0] - qq[2,0]',
        )
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[10].lower, 0.0)
        self.assertEqual(m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[10].upper, 0.0)
        self.assertEqual(
            m_kaug._SENSITIVITY_TOOLBOX_DATA.paramConst[10].body.to_string(),
            '_SENSITIVITY_TOOLBOX_DATA.aa - aa',
        )

    # Test example `parameter_kaug.py`
    @unittest.skipIf(not opt_kaug.available(False), "k_aug is not available")
    @unittest.skipIf(not opt_dotsens.available(False), "dot_sens is not available")
    def test_parameter_example_kaug(self):
        d = param_kaug_ex.run_example()

        d_correct = {
            'eta1': 4.5,
            'eta2': 1.0,
            'x1_init': 0.15,
            'x2_init': 0.15,
            'x3_init': 0.0,
            'eta1_pert': 4.0,
            'eta2_pert': 1.0,
            'x1_pert': 0.3333333,
            'x2_pert': 0.6666667,
            'x3_pert': 0.0,
            'cost_pert': 0.55555556,
        }

        for k in d_correct.keys():
            # Check each element of the 'correct' dictionary against the returned
            # dictionary to 3 decimal places
            self.assertAlmostEqual(d[k], d_correct[k], 3)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
#  SOS tests
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep
import pyomo.common.unittest as unittest
from pyomo.environ import ConcreteModel, AbstractModel, SOSConstraint, Var, Set


class TestErrors(unittest.TestCase):
    def test_arg1(self):
        M = ConcreteModel()
        try:
            M.c = SOSConstraint()
            self.fail("Expected TypeError")
        except TypeError:
            pass

    def test_arg2(self):
        M = ConcreteModel()
        M.v = Var()
        try:
            M.c = SOSConstraint(var=M.v, sos=1, level=1)
            self.fail("Expected TypeError")
        except TypeError:
            pass

    def test_arg3(self):
        M = ConcreteModel()
        M.v = Var()
        try:
            M.c = SOSConstraint(var=M.v)
            self.fail("Expected TypeError")
        except TypeError:
            pass

    def test_negative_weights(self):
        M = ConcreteModel()
        M.v = Var()
        try:
            M.c = SOSConstraint(var=M.v, weights={None: -1}, sos=1)
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_ordered(self):
        M = ConcreteModel()
        M.v = Var({1, 2, 3})
        try:
            M.c = SOSConstraint(var=M.v, sos=2)
            self.fail("Expected ValueError")
        except ValueError:
            pass
        M = ConcreteModel()
        M.s = Set(initialize=[1, 2, 3], ordered=True)
        M.v = Var(M.s)
        M.c = SOSConstraint(var=M.v, sos=2)


class TestSimple(unittest.TestCase):
    def setUp(self):
        #
        # Create Model
        #
        self.M = ConcreteModel()

    def tearDown(self):
        self.M = None

    def test_num_vars(self):
        # Test the number of variables
        self.M.x = Var([1, 2, 3])
        self.M.c = SOSConstraint(var=self.M.x, sos=1)
        self.assertEqual(self.M.c.num_variables(), 3)

    def test_level(self):
        # Test level property
        self.M.x = Var([1, 2, 3])
        self.M.c = SOSConstraint(var=self.M.x, sos=1)
        self.assertEqual(self.M.c.level, 1)
        self.M.c.level = 2
        self.assertEqual(self.M.c.level, 2)
        try:
            self.M.c.level = -1
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_get_variables(self):
        # Test that you get the correct variables
        self.M.x = Var([1, 2, 3])
        self.M.c = SOSConstraint(var=self.M.x, sos=1)
        self.assertEqual(
            set(id(v) for v in self.M.c.get_variables()),
            set(id(v) for v in self.M.x.values()),
        )


class TestExamples(unittest.TestCase):
    def test1(self):
        M = ConcreteModel()
        M.x = Var(range(20))
        M.c = SOSConstraint(var=M.x, sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c.get_items()),
            set((M.x[i].name, i + 1) for i in range(20)),
        )

    def test2(self):
        # Use an index set, which is a subset of M.x.index_set()
        M = ConcreteModel()
        M.x = Var(range(20))
        M.c = SOSConstraint(var=M.x, index=list(range(10)), sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c.get_items()),
            set((M.x[i].name, i + 1) for i in range(10)),
        )

    def test3(self):
        # User-specified weights
        w = {1: 10, 2: 2, 3: 30}
        M = ConcreteModel()
        M.x = Var([1, 2, 3])
        M.c = SOSConstraint(var=M.x, weights=w, sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c.get_items()),
            set((M.x[i].name, w[i]) for i in [1, 2, 3]),
        )

    def test4(self):
        # User-specified weights
        w = {1: 10, 2: 2, 3: 30}

        def rule(model):
            return list(M.x[i] for i in M.x), [10, 2, 30]

        M = ConcreteModel()
        M.x = Var([1, 2, 3], dense=True)
        M.c = SOSConstraint(rule=rule, sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c.get_items()),
            set((M.x[i].name, w[i]) for i in [1, 2, 3]),
        )

    def test10(self):
        M = ConcreteModel()
        M.x = Var([1, 2, 3])
        M.c = SOSConstraint([0, 1], var=M.x, sos=1, index={0: [1, 2], 1: [2, 3]})
        self.assertEqual(
            set((v.name, w) for v, w in M.c[0].get_items()),
            set((M.x[i].name, i) for i in [1, 2]),
        )
        self.assertEqual(
            set((v.name, w) for v, w in M.c[1].get_items()),
            set((M.x[i].name, i - 1) for i in [2, 3]),
        )

    def test11(self):
        w = {1: 10, 2: 2, 3: 30}
        M = ConcreteModel()
        M.x = Var([1, 2, 3], dense=True)
        M.c = SOSConstraint(
            [0, 1], var=M.x, weights=w, sos=1, index={0: [1, 2], 1: [2, 3]}
        )
        self.assertEqual(
            set((v.name, w) for v, w in M.c[0].get_items()),
            set((M.x[i].name, w[i]) for i in [1, 2]),
        )
        self.assertEqual(
            set((v.name, w) for v, w in M.c[1].get_items()),
            set((M.x[i].name, w[i]) for i in [2, 3]),
        )

    def test12(self):
        def rule(model, i):
            if i == 0:
                return list(M.x[i] for i in M.x), [10, 2, 30]
            else:
                return list(M.x[i] for i in M.x), [1, 20, 3]

        w = {0: {1: 10, 2: 2, 3: 30}, 1: {1: 1, 2: 20, 3: 3}}
        M = ConcreteModel()
        M.x = Var([1, 2, 3], dense=True)
        M.c = SOSConstraint([0, 1], rule=rule, sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c[0].get_items()),
            set((M.x[i].name, w[0][i]) for i in [1, 2, 3]),
        )
        self.assertEqual(
            set((v.name, w) for v, w in M.c[1].get_items()),
            set((M.x[i].name, w[1][i]) for i in [1, 2, 3]),
        )

    def test13(self):
        I = {0: [1, 2], 1: [2, 3]}
        M = ConcreteModel()
        M.x = Var([1, 2, 3], dense=True)
        M.c = SOSConstraint([0, 1], var=M.x, index=I, sos=1)
        self.assertEqual(
            set((v.name, w) for v, w in M.c[0].get_items()),
            set((M.x[i].name, i) for i in I[0]),
        )
        self.assertEqual(
            set((v.name, w) for v, w in M.c[1].get_items()),
            set((M.x[i].name, i - 1) for i in I[1]),
        )

    def test14(self):
        def rule(model, i):
            if i == 0:
                return SOSConstraint.Skip
            else:
                return list(M.x[i] for i in M.x), [1, 20, 3]

        w = {0: {1: 10, 2: 2, 3: 30}, 1: {1: 1, 2: 20, 3: 3}}
        M = ConcreteModel()
        M.x = Var([1, 2, 3], dense=True)
        M.c = SOSConstraint([0, 1], rule=rule, sos=1)
        self.assertEqual(list(M.c.keys()), [1])
        self.assertEqual(
            set((v.name, w) for v, w in M.c[1].get_items()),
            set((M.x[i].name, w[1][i]) for i in [1, 2, 3]),
        )

    def test_abstract_index(self):
        model = AbstractModel()
        model.A = Set(initialize=[0])
        model.B = Set(initialize=[1])
        model.C = model.A | model.B
        M = ConcreteModel()
        M.x = Var([1, 2, 3])
        M.c = SOSConstraint(model.C, var=M.x, sos=1, index={0: [1, 2], 1: [2, 3]})


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.contrib.pynumero.examples.callback.reactor_design import model as m
import logging

"""
This example uses an iteration callback to print the values
of the constraint residuals at each iteration of the CyIpopt
solver
"""


def iteration_callback(
    nlp,
    alg_mod,
    iter_count,
    obj_value,
    inf_pr,
    inf_du,
    mu,
    d_norm,
    regularization_size,
    alpha_du,
    alpha_pr,
    ls_trials,
):
    logger = logging.getLogger('pyomo')
    constraint_names = nlp.constraint_names()
    residuals = nlp.evaluate_constraints()
    logger.info('      ...Residuals for iteration {}'.format(iter_count))
    for i, nm in enumerate(constraint_names):
        logger.info('      ...{}: {}'.format(nm, residuals[i]))


def main():
    solver = pyo.SolverFactory('cyipopt')
    status, nlp = solver.solve(
        m, tee=False, return_nlp=True, intermediate_callback=iteration_callback
    )


if __name__ == '__main__':
    logging.getLogger('pyomo').setLevel(logging.INFO)
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import pandas as pd
from os.path import join, abspath, dirname
import pyomo.environ as pyo
import pyomo.contrib.parmest.parmest as parmest
from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
    ReactorDesignExperiment,
)


class MultisensorReactorDesignExperiment(ReactorDesignExperiment):

    def finalize_model(self):

        m = self.model

        # Experiment inputs values
        m.sv = self.data_i['sv']
        m.caf = self.data_i['caf']

        # Experiment output values
        m.ca = (self.data_i['ca1'] + self.data_i['ca2'] + self.data_i['ca3']) * (1 / 3)
        m.cb = self.data_i['cb']
        m.cc = (self.data_i['cc1'] + self.data_i['cc2']) * (1 / 2)
        m.cd = self.data_i['cd']

        return m

    def label_model(self):

        m = self.model

        m.experiment_outputs = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.experiment_outputs.update(
            [
                (m.ca, [self.data_i['ca1'], self.data_i['ca2'], self.data_i['ca3']]),
                (m.cb, [self.data_i['cb']]),
                (m.cc, [self.data_i['cc1'], self.data_i['cc2']]),
                (m.cd, [self.data_i['cd']]),
            ]
        )

        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters.update(
            (k, pyo.ComponentUID(k)) for k in [m.k1, m.k2, m.k3]
        )

        return m


def main():
    # Parameter estimation using multisensor data

    # Read in data
    file_dirname = dirname(abspath(str(__file__)))
    file_name = abspath(join(file_dirname, "reactor_data_multisensor.csv"))
    data = pd.read_csv(file_name)

    # Create an experiment list
    exp_list = []
    for i in range(data.shape[0]):
        exp_list.append(MultisensorReactorDesignExperiment(data, i))

    # Define sum of squared error
    def SSE_multisensor(model):
        expr = 0
        for y, y_hat in model.experiment_outputs.items():
            num_outputs = len(y_hat)
            for i in range(num_outputs):
                expr += ((y - y_hat[i]) ** 2) * (1 / num_outputs)
        return expr

    # View one model
    # exp0_model = exp_list[0].get_labeled_model()
    # exp0_model.pprint()
    # print(SSE_multisensor(exp0_model))

    pest = parmest.Estimator(exp_list, obj_function=SSE_multisensor)
    obj, theta = pest.theta_est()
    print(obj)
    print(theta)


if __name__ == "__main__":
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import json
import logging
import os, os.path
from glob import glob

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
    scipy_available,
    matplotlib,
    matplotlib_available,
)

# Set matplotlib backend for non-interactive use (for CI testing purposes)
if matplotlib_available:
    matplotlib.use("Agg")

from pyomo.common.fileutils import this_file_dir
import pyomo.common.unittest as unittest

if not (numpy_available and scipy_available):
    raise unittest.SkipTest("Pyomo.DoE needs scipy and numpy to run tests")

if scipy_available:
    from pyomo.contrib.doe import DesignOfExperiments
    from pyomo.contrib.doe.examples.reactor_experiment import ReactorExperiment
    from pyomo.contrib.doe.examples.reactor_example import (
        ReactorExperiment as FullReactorExperiment,
        run_reactor_doe,
    )
    from pyomo.contrib.doe.tests.experiment_class_example_flags import (
        RooneyBieglerExperimentBad,
    )
    from pyomo.contrib.parmest.examples.rooney_biegler.rooney_biegler import (
        RooneyBieglerExperiment,
    )
from pyomo.contrib.doe.utils import rescale_FIM
from pyomo.contrib.doe.examples.rooney_biegler_doe_example import run_rooney_biegler_doe

import pyomo.environ as pyo

from pyomo.opt import SolverFactory

ipopt_available = SolverFactory("ipopt").available()
k_aug_available = SolverFactory("k_aug", solver_io="nl", validate=False)

currdir = this_file_dir()
file_path = os.path.join(currdir, "..", "examples", "result.json")

with open(file_path) as f:
    data_ex = json.load(f)
data_ex["control_points"] = {float(k): v for k, v in data_ex["control_points"].items()}


def get_rooney_biegler_data():
    """Get Rooney-Biegler experiment data for testing."""
    # Create a simple data point for Rooney-Biegler model
    # This must be a pandas Series with 'hour' and 'y' columns
    # Use data from the tested Rooney-Biegler dataset
    data = pd.DataFrame(data=[[5, 15.6]], columns=['hour', 'y'])
    return data.iloc[0]


def get_rooney_biegler_experiment():
    """Get a fresh RooneyBieglerExperiment instance for testing.

    Creates a new experiment instance to ensure test isolation.
    Each test gets its own instance to avoid state sharing.
    """
    return RooneyBieglerExperiment(
        data=get_rooney_biegler_data(),
        theta={'asymptote': 15, 'rate_constant': 0.5},
        measure_error=0.1,
    )


def get_FIM_Q_L(doe_obj=None):
    """
    Helper function to retrieve results to compare.

    """
    model = doe_obj.model

    n_param = doe_obj.n_parameters
    n_y = doe_obj.n_experiment_outputs

    FIM_vals = [
        pyo.value(model.fim[i, j])
        for i in model.parameter_names
        for j in model.parameter_names
    ]
    if hasattr(model, "L"):
        L_vals = [
            pyo.value(model.L[i, j])
            for i in model.parameter_names
            for j in model.parameter_names
        ]
    else:
        L_vals = [[0] * n_param] * n_param
    Q_vals = [
        pyo.value(model.sensitivity_jacobian[i, j])
        for i in model.output_names
        for j in model.parameter_names
    ]
    sigma_inv = [
        1 / v**2 for k, v in model.scenario_blocks[0].measurement_error.items()
    ]
    param_vals = np.array(
        [[v for k, v in model.scenario_blocks[0].unknown_parameters.items()]]
    )

    FIM_vals_np = np.array(FIM_vals).reshape((n_param, n_param))

    for i in range(n_param):
        for j in range(n_param):
            if j < i:
                FIM_vals_np[j, i] = FIM_vals_np[i, j]

    L_vals_np = np.array(L_vals).reshape((n_param, n_param))
    Q_vals_np = np.array(Q_vals).reshape((n_y, n_param))

    sigma_inv_np = np.zeros((n_y, n_y))

    for ind, v in enumerate(sigma_inv):
        sigma_inv_np[ind, ind] = v

    return FIM_vals_np, Q_vals_np, L_vals_np, sigma_inv_np


def get_standard_args(experiment, fd_method, obj_used):
    args = {}
    args['experiment'] = experiment
    args['fd_formula'] = fd_method
    args['step'] = 1e-3
    args['objective_option'] = obj_used
    args['scale_constant_value'] = 1
    args['scale_nominal_param_value'] = True
    args['prior_FIM'] = None
    args['jac_initial'] = None
    args['fim_initial'] = None
    args['L_diagonal_lower_bound'] = 1e-7
    # Make solver object with good linear subroutines
    solver = SolverFactory("ipopt")
    solver.options["linear_solver"] = "ma57"
    solver.options["halt_on_ampl_error"] = "yes"
    solver.options["max_iter"] = 3000
    args['solver'] = solver
    args['tee'] = False
    args['get_labeled_model_args'] = None
    args['_Cholesky_option'] = True
    args['_only_compute_fim_lower'] = True
    return args


@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
@unittest.skipIf(not numpy_available, "Numpy is not available")
@unittest.skipIf(not scipy_available, "scipy is not available")
class TestRooneyBieglerExampleSolving(unittest.TestCase):
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_rooney_biegler_fd_central_solve(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.run_doe()

        # assert model solves
        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        # assert that Q, F, and L are the same.
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)

        # Since Trace is used, no comparison for FIM and L.T @ L

        # Make sure FIM and Q.T @ sigma_inv @ Q are close (alternate definition of FIM)
        self.assertTrue(np.all(np.isclose(FIM, Q.T @ sigma_inv @ Q)))

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_rooney_biegler_fd_forward_solve(self):
        fd_method = "forward"
        obj_used = "zero"

        # Use RooneyBiegler for algorithm validation (faster)
        # Use hour=7 for better FIM conditioning with zero objective
        data_point = pd.DataFrame({'hour': [7.0], 'y': [19.8]}).iloc[0]

        experiment = RooneyBieglerExperiment(
            data=data_point,
            theta={'asymptote': 15, 'rate_constant': 0.5},
            measure_error=0.1,
        )

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        # Add prior FIM to avoid singularity with zero objective
        # This follows the pattern in rooney_biegler_doe_example.py
        doe_obj_prior = DesignOfExperiments(**DoE_args)
        prior_FIM = doe_obj_prior.compute_FIM()
        DoE_args['prior_FIM'] = prior_FIM

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.run_doe()

        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        # assert that Q, F, and L are the same.
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)

        # Since Trace is used, no comparison for FIM and L.T @ L

        # Note: When using prior_FIM, the relationship FIM = Q.T @ sigma_inv @ Q + prior_FIM
        self.assertTrue(np.all(np.isclose(FIM, Q.T @ sigma_inv @ Q + prior_FIM)))

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_rooney_biegler_fd_backward_solve(self):
        fd_method = "backward"
        obj_used = "pseudo_trace"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.run_doe()

        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        # assert that Q, F, and L are the same.
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)

        # Since Trace is used, no comparison for FIM and L.T @ L

        # Make sure FIM and Q.T @ sigma_inv @ Q are close (alternate definition of FIM)
        self.assertTrue(np.all(np.isclose(FIM, Q.T @ sigma_inv @ Q)))

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_rooney_biegler_obj_det_solve(self):
        fd_method = "central"
        obj_used = "determinant"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)
        DoE_args["scale_nominal_param_value"] = (
            False  # Vanilla determinant solve needs this
        )
        DoE_args["_Cholesky_option"] = False
        DoE_args["_only_compute_fim_lower"] = False

        doe_obj = DesignOfExperiments(**DoE_args)

        # Increase numerical performance by adding a prior
        prior_FIM = doe_obj.compute_FIM()
        doe_obj.prior_FIM = prior_FIM

        doe_obj.run_doe()

        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        expected_design = 9.999213890476453
        actual_design = doe_obj.results["Experiment Design"][0]
        self.assertAlmostEqual(actual_design, expected_design, places=3)

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_rooney_biegler_obj_cholesky_solve(self):
        fd_method = "central"
        obj_used = "determinant"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        # Add prior FIM for better numerical conditioning
        # This follows the pattern in rooney_biegler_doe_example.py
        doe_obj_prior = DesignOfExperiments(**DoE_args)
        prior_FIM = doe_obj_prior.compute_FIM()
        DoE_args['prior_FIM'] = prior_FIM

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.run_doe()

        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        # assert that Q, F, and L are the same.
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)

        # Since Cholesky is used, there is comparison for FIM and L.T @ L
        self.assertTrue(np.all(np.isclose(FIM, L @ L.T)))

        # Note: When using prior_FIM, the relationship FIM = Q.T @ sigma_inv @ Q + prior_FIM
        self.assertTrue(np.all(np.isclose(FIM, Q.T @ sigma_inv @ Q + prior_FIM)))

    def DISABLE_test_reactor_obj_cholesky_solve_bad_prior(self):
        # [10/2025] This test has been disabled because it frequently
        # (and randomly) returns "infeasible" when run on Windows.
        from pyomo.contrib.doe.doe import _SMALL_TOLERANCE_DEFINITENESS

        fd_method = "central"
        obj_used = "determinant"

        experiment = FullReactorExperiment(data_ex, 10, 3)

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        # Specify a prior that is slightly negative definite
        # Because it is less than the tolerance, it should be
        # adjusted to be positive definite
        # No error should be thrown
        DoE_args["prior_FIM"] = -(_SMALL_TOLERANCE_DEFINITENESS / 100) * np.eye(4)

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.run_doe()

        self.assertEqual(doe_obj.results["Solver Status"], "ok")

        # assert that Q, F, and L are the same.
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)

        # Since Cholesky is used, there is comparison for FIM and L.T @ L
        self.assertTrue(np.all(np.isclose(FIM, L @ L.T)))

        # Make sure FIM and Q.T @ sigma_inv @ Q are close (alternate definition of FIM)
        self.assertTrue(np.all(np.isclose(FIM, Q.T @ sigma_inv @ Q)))

    # This test ensure that compute FIM runs without error using the
    # `sequential` option with central finite differences
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_compute_FIM_seq_centr(self):
        fd_method = "central"
        obj_used = "pseudo_trace"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        expected_FIM = np.array(
            [[18957.7788694, 4238.27606876], [4238.27606876, 947.52577076]]
        )

        self.assertTrue(
            np.all(np.isclose(doe_obj.compute_FIM(method="sequential"), expected_FIM))
        )

    # This test ensure that compute FIM runs without error using the
    # `sequential` option with forward finite differences
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_compute_FIM_seq_forward(self):
        fd_method = "forward"
        obj_used = "pseudo_trace"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.compute_FIM(method="sequential")

    # This test ensure that compute FIM runs without error using the
    # `kaug` option. kaug computes the FIM directly so no finite difference
    # scheme is needed.
    @unittest.skipIf(not scipy_available, "Scipy is not available")
    @unittest.skipIf(
        not k_aug_available.available(False), "The 'k_aug' command is not available"
    )
    def test_compute_FIM_kaug(self):
        fd_method = "forward"
        obj_used = "determinant"

        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        expected_FIM = np.array(
            [[18957.7788694, 4238.27606876], [4238.27606876, 947.52577076]]
        )

        self.assertTrue(
            np.all(np.isclose(doe_obj.compute_FIM(method="kaug"), expected_FIM))
        )

    # This test ensure that compute FIM runs without error using the
    # `sequential` option with backward finite differences
    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_compute_FIM_seq_backward(self):
        fd_method = "backward"
        obj_used = "pseudo_trace"

        # Use RooneyBiegler for algorithm validation (faster)
        experiment = get_rooney_biegler_experiment()

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.compute_FIM(method="sequential")

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_reactor_grid_search(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = FullReactorExperiment(data_ex, 10, 3)

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        # Reduce grid from 3x3 to 2x2 for performance
        design_ranges = {"CA[0]": [1, 5, 2], "T[0]": [300, 700, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        # Check to make sure the lengths of the inputs
        # in results object are indeed correct
        CA_vals = doe_obj.fim_factorial_results["CA[0]"]
        T_vals = doe_obj.fim_factorial_results["T[0]"]

        # assert length is correct (2x2 = 4 evaluations)
        self.assertTrue((len(CA_vals) == 4) and (len(T_vals) == 4))
        self.assertTrue((len(set(CA_vals)) == 2) and (len(set(T_vals)) == 2))

        # assert unique values are correct
        self.assertTrue(
            (set(CA_vals).issuperset(set([1, 5])))
            and (set(T_vals).issuperset(set([300, 700])))
        )

    def test_rescale_FIM(self):
        fd_method = "central"
        obj_used = "determinant"

        experiment = FullReactorExperiment(data_ex, 10, 3)

        # With parameter scaling
        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        # Without parameter scaling
        DoE_args2 = get_standard_args(experiment, fd_method, obj_used)
        DoE_args2["scale_nominal_param_value"] = False

        doe_obj2 = DesignOfExperiments(**DoE_args2)
        # Run both problems
        doe_obj.run_doe()
        doe_obj2.run_doe()

        # Extract FIM values
        FIM, Q, L, sigma_inv = get_FIM_Q_L(doe_obj=doe_obj)
        FIM2, Q2, L2, sigma_inv2 = get_FIM_Q_L(doe_obj=doe_obj2)

        # Get rescaled FIM from the scaled version
        param_vals = np.array(
            [
                [
                    v
                    for k, v in doe_obj.model.scenario_blocks[
                        0
                    ].unknown_parameters.items()
                ]
            ]
        )

        resc_FIM = rescale_FIM(FIM, param_vals)

        # Compare scaled and rescaled values
        self.assertTrue(np.all(np.isclose(FIM2, resc_FIM)))

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_reactor_solve_bad_model(self):
        fd_method = "central"
        obj_used = "determinant"

        # Use RooneyBiegler bad example (faster than reactor bad example)
        experiment = RooneyBieglerExperimentBad(
            data=get_rooney_biegler_data(),
            theta={'asymptote': 15, 'rate_constant': 0.5},
            measure_error=0.1,
        )

        DoE_args = get_standard_args(experiment, fd_method, obj_used)

        doe_obj = DesignOfExperiments(**DoE_args)

        with self.assertRaisesRegex(
            RuntimeError,
            "Model from experiment did not solve appropriately. "
            "Make sure the model is well-posed.",
        ):
            doe_obj.run_doe()

    @unittest.skipIf(not pandas_available, "pandas is not available")
    def test_reactor_grid_search_bad_model(self):
        fd_method = "central"
        obj_used = "determinant"

        # Use RooneyBiegler bad example (faster than reactor bad example)
        experiment = RooneyBieglerExperimentBad(
            data=get_rooney_biegler_data(),
            theta={'asymptote': 15, 'rate_constant': 0.5},
            measure_error=0.1,
        )

        DoE_args = get_standard_args(experiment, fd_method, obj_used)
        DoE_args["logger_level"] = logging.ERROR

        doe_obj = DesignOfExperiments(**DoE_args)

        # Use simpler design ranges for RooneyBiegler
        design_ranges = {"hour": [1, 5, 2]}

        doe_obj.compute_FIM_full_factorial(
            design_ranges=design_ranges, method="sequential"
        )

        # Check to make sure the lengths of the inputs in results object are indeed correct
        hour_vals = doe_obj.fim_factorial_results["hour"]

        # assert length is correct
        self.assertTrue(len(hour_vals) == 2)
        self.assertTrue(len(set(hour_vals)) == 2)

        # assert unique values are correct
        self.assertTrue(set(hour_vals).issuperset(set([1, 5])))


@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
@unittest.skipIf(not numpy_available, "Numpy is not available")
class TestDoe(unittest.TestCase):
    def test_doe_full_factorial(self):
        log10_D_opt_expected = [
            11.77343778527225,
            13.137792359064383,
            13.182167857699808,
            14.54652243150573,
        ]

        log10_A_opt_expected = [
            5.59357268009304,
            5.613318615148643,
            5.945755198204368,
            5.965501133259909,
        ]

        log10_E_opt_expected = [
            0.27981268741620413,
            1.3086595026369012,
            0.6319952055040333,
            1.6608420207466377,
        ]

        log10_ME_opt_expected = [
            5.221185311075697,
            4.244741560076784,
            5.221185311062606,
            4.244741560083524,
        ]

        eigval_min_expected = [
            1.9046390638130666,
            20.354456134677426,
            4.285437893696232,
            45.797526302234304,
        ]

        eigval_max_expected = [
            316955.2855492114,
            357602.92523637977,
            713149.3924857995,
            804606.58178139165,
        ]

        det_FIM_expected = [
            593523317093.4525,
            13733851875566.766,
            15211353450350.424,
            351983602166961.56,
        ]

        trace_FIM_expected = [
            392258.78617108597,
            410505.1549241871,
            882582.2688850109,
            923636.598578955,
        ]
        ff = run_reactor_doe(
            n_points_for_design=2,
            compute_FIM_full_factorial=False,
            plot_factorial_results=False,
            run_optimal_doe=False,
        )
        ff.compute_FIM_full_factorial(
            design_ranges={"CA[0]": [1, 1.5, 2], "T[0]": [350, 400, 2]}
        )

        ff_results = ff.fim_factorial_results

        self.assertStructuredAlmostEqual(
            ff_results["log10 D-opt"], log10_D_opt_expected, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            ff_results["log10 pseudo A-opt"], log10_A_opt_expected, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            ff_results["log10 E-opt"], log10_E_opt_expected, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            ff_results["log10 ME-opt"], log10_ME_opt_expected, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            ff_results["eigval_min"], eigval_min_expected, abstol=1e-4
        )
        # abstol of 1e-4 removed for the following values as
        # their non-log values are large (e.g., >1e10)
        self.assertStructuredAlmostEqual(ff_results["eigval_max"], eigval_max_expected)
        self.assertStructuredAlmostEqual(ff_results["det_FIM"], det_FIM_expected)
        self.assertStructuredAlmostEqual(ff_results["trace_FIM"], trace_FIM_expected)

    @unittest.skipUnless(pandas_available, "test requires pandas")
    def test_doe_A_optimality(self):
        A_opt_value_expected = -2.2364242059539663
        A_opt_design_value_expected = 9.999955457176451

        A_opt_res = run_rooney_biegler_doe(optimization_objective="trace")
        A_opt_value = A_opt_res["optimization"]["value"]
        A_opt_design_value = A_opt_res["optimization"]["design"][0]

        self.assertAlmostEqual(A_opt_value, A_opt_value_expected, places=2)
        # print("A optimal design value:", A_opt_design_value)
        self.assertAlmostEqual(
            A_opt_design_value, A_opt_design_value_expected, places=2
        )


class TestRooneyBieglerExample(unittest.TestCase):
    @unittest.skipUnless(pandas_available, "test requires pandas")
    @unittest.skipUnless(ipopt_available, "test requires ipopt")
    def test_rooney_biegler_doe_example(self):
        """
        Tests the Design of Experiments (DoE) functionality, including
        plotting logic when matplotlib is available, without displaying GUI windows.
        """
        file_prefix = "rooney_biegler"

        # Cleanup function for generated files
        def cleanup_file():
            generated_files = glob(f"{file_prefix}_*.png")
            for f in generated_files:
                try:
                    os.remove(f)
                except OSError:
                    pass

        self.addCleanup(cleanup_file)

        # Run with draw_factorial_figure conditional on matplotlib availability
        # Test D-optimality
        results_D = run_rooney_biegler_doe(
            optimization_objective="determinant",
            compute_FIM_full_factorial=True,
            draw_factorial_figure=matplotlib_available,
            design_range={'hour': [0, 10, 3]},
            tee=False,
        )

        # Test A-optimality
        results_A = run_rooney_biegler_doe(
            optimization_objective="trace",
            compute_FIM_full_factorial=False,
            draw_factorial_figure=False,
            design_range={'hour': [0, 10, 3]},
            tee=False,
        )

        # Assertions for Numerical Results
        self.assertEqual("determinant", results_D["optimization"]["objective_type"])
        self.assertEqual("trace", results_A["optimization"]["objective_type"])

        # Test D-optimality optimization results
        D_opt_value_expected = 6.864794717802814
        D_opt_design_value_expected = 10.0  # approximately 9.999999472662282

        D_opt_value = results_D["optimization"]["value"]
        D_opt_design_value = results_D["optimization"]["design"][0]

        self.assertAlmostEqual(D_opt_value, D_opt_value_expected, places=4)
        self.assertAlmostEqual(
            D_opt_design_value, D_opt_design_value_expected, places=4
        )

        # Test A-optimality optimization results
        A_opt_value_expected = -2.236424205953928
        A_opt_design_value_expected = 10.0  # approximately 9.999955457176451

        A_opt_value = results_A["optimization"]["value"]
        A_opt_design_value = results_A["optimization"]["design"][0]

        self.assertAlmostEqual(A_opt_value, A_opt_value_expected, places=4)
        self.assertAlmostEqual(
            A_opt_design_value, A_opt_design_value_expected, places=4
        )

        # Assertions for Full Factorial Results
        self.assertIn("results_dict", results_D)
        results_dict = results_D["results_dict"]
        self.assertIsInstance(results_dict, dict)
        self.assertGreater(len(results_dict), 0, "results_dict should not be empty")

        # Expected values for design_range={'hour': [0, 10, 3]}
        # These are the 3 data points from the full factorial grid
        expected_log10_D_opt = [6.583798747893548, 6.691228337572129, 6.864794726228617]
        expected_log10_A_opt = [
            -1.9574859220185146,
            -2.0268526846104975,
            -2.236424954559946,
        ]
        expected_log10_pseudo_A_opt = [
            4.62631282587503,
            4.6643756529616285,
            4.628369771668666,
        ]
        expected_log10_E_opt = [
            1.9584199467177335,
            2.0278567624056967,
            2.2381970919918426,
        ]
        expected_log10_ME_opt = [
            2.666958854458125,
            2.6355148127607713,
            2.3884005422449364,
        ]

        # Verify structure and values using assertStructuredAlmostEqual
        self.assertStructuredAlmostEqual(
            results_dict["log10 D-opt"], expected_log10_D_opt, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            results_dict["log10 A-opt"], expected_log10_A_opt, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            results_dict["log10 pseudo A-opt"], expected_log10_pseudo_A_opt, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            results_dict["log10 E-opt"], expected_log10_E_opt, abstol=1e-4
        )
        self.assertStructuredAlmostEqual(
            results_dict["log10 ME-opt"], expected_log10_ME_opt, abstol=1e-4
        )

        # Plot-related assertions only when matplotlib is available
        if matplotlib_available:
            # Check that draw_factorial_figure actually created the file
            expected_d_plot = f"{file_prefix}_D_opt.png"
            self.assertTrue(
                os.path.exists(expected_d_plot),
                f"Expected plot file '{expected_d_plot}' was not created.",
            )


@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
@unittest.skipIf(not numpy_available, "Numpy is not available")
@unittest.skipIf(not pandas_available, "Pandas is not available")
@unittest.skipIf(not matplotlib_available, "Matplotlib is not available")
class TestDoEFactorialFigure(unittest.TestCase):
    def test_doe_1D_plotting_function(self):
        # For 1D plotting we will use the Rooney-Biegler example in parmest/examples
        plt = matplotlib.pyplot
        """
        Test that the plotting function executes without error and
        creates a matplotlib figure. We do NOT test visual correctness.
        """

        # File prefix for saved plots
        # Define prefixes for the two runs
        prefix_linear = "rooney_linear"
        prefix_log = "rooney_log"

        # Clean up any existing plot files from test runs
        def cleanup_files():
            files_to_remove = glob("rooney_*.png")
            for f in files_to_remove:
                try:
                    os.remove(f)
                except OSError:
                    pass
            plt.close('all')

        self.addCleanup(cleanup_files)

        fd_method = "central"
        obj_used = "trace"

        experiment = run_rooney_biegler_doe()["experiment"]

        DoE_args = get_standard_args(experiment, fd_method, obj_used)
        doe_obj = DesignOfExperiments(**DoE_args)

        doe_obj.compute_FIM_full_factorial(design_ranges={'hour': [0, 10, 1]})

        # Call the plotting function for linear scale
        doe_obj.draw_factorial_figure(
            sensitivity_design_variables=['hour'],
            fixed_design_variables={},
            log_scale=False,
            figure_file_name=prefix_linear,
        )

        # Call the plotting function for log scale
        doe_obj.draw_factorial_figure(
            sensitivity_design_variables=['hour'],
            fixed_design_variables={},
            log_scale=True,
            figure_file_name=prefix_log,
        )

        # Verify that the linear scale plots were also created
        # Check that we found exactly 5 files (A, D, E, ME, pseudo_A)
        expected_plot_linear = glob(f"{prefix_linear}*.png")
        self.assertEqual(
            len(expected_plot_linear),
            5,
            f"Expected 5 plot files, but found {len(expected_plot_linear)}. Files found: {expected_plot_linear}",
        )

        # Verify that the log scale plots were also created
        expected_plot_log = glob(f"{prefix_log}*.png")
        self.assertEqual(
            len(expected_plot_log),
            5,
            f"Expected 5 plot files, but found {len(expected_plot_log)}. Files found: {expected_plot_log}",
        )

    def test_doe_2D_plotting_function(self):
        # For 2D plotting we will use the Rooney-Biegler example in doe/examples
        plt = matplotlib.pyplot

        # File prefix for saved plots
        prefix_linear = "reactor_linear"
        prefix_log = "reactor_log"

        # Clean up any existing plot files from test runs
        def cleanup_files():
            files_to_remove = glob("reactor_*.png")
            for f in files_to_remove:
                try:
                    os.remove(f)
                except OSError:
                    pass
            plt.close('all')

        self.addCleanup(cleanup_files)

        # Run the reactor example
        run_reactor_doe(
            n_points_for_design=1,
            compute_FIM_full_factorial=True,
            plot_factorial_results=True,
            figure_file_name=prefix_linear,
            log_scale=False,
            run_optimal_doe=False,
        )

        # Verify that the linear scale plots were also created
        # Check that we found exactly 5 files (A, D, E, ME, pseudo_A)
        expected_plot_linear = glob(f"{prefix_linear}*.png")
        self.assertTrue(
            len(expected_plot_linear) == 5,
            f"Expected 5 plot files, but found {len(expected_plot_linear)}. Files found: {expected_plot_linear}",
        )

        # Run the reactor example with log scale
        run_reactor_doe(
            n_points_for_design=1,
            compute_FIM_full_factorial=True,
            plot_factorial_results=True,
            figure_file_name=prefix_log,
            log_scale=True,
            run_optimal_doe=False,
        )

        # Verify that the log scale plots were also created
        # Check that we found exactly 5 files (A, D, E, ME, pseudo_A)
        expected_plot_log = glob(f"{prefix_log}*.png")
        self.assertTrue(
            len(expected_plot_log) == 5,
            f"Expected 5 plot files, but found {len(expected_plot_log)}. Files found: {expected_plot_log}",
        )


if __name__ == "__main__":
    unittest.main()
